from __future__ import division

import math
import numpy
import os

import dill
#from klepto.safe import inf_cache as memoized
from klepto import inf_cache as memoized

from klepto.archives import dir_archive
from klepto.keymaps import picklemap

dumps = picklemap(serializer='dill')


def random_random_3d(bins,npoints,boxsize):
    ndens = npoints / boxsize**3.
    bin_shift = numpy.roll(bins,1)
    bin_shift[0] = 0.

    return npoints * (4./3.) * math.pi * ndens * (bins**3 - bin_shift**3) / 2.
    # 1/2 factor comes from autocorrelation...

def random_random(bins,pimax,npoints,boxsize):
    ndens = npoints / boxsize**3.
    bin_shift = numpy.roll(bins,1)
    bin_shift[0] = 0.

    return npoints * math.pi * ndens * (bins**2 - bin_shift**2) * pimax
    # factor of 2*pimax / 2 = pimax, since 1/2 factor comes from autocorrelation...


@memoized(maxsize=None,cache=dir_archive('wp_cache',serialized=True,cached=True),keymap=dumps)
def data_data(bins,pimax,array_galaxy,boxsize):
    return numpy.zeros(bins.shape)

@memoized(maxsize=None,cache=dir_archive('xsi_cache',serialized=True,cached=True),keymap=dumps)
def data_data_3d(bins,array_galaxy,boxsize):
    return numpy.zeros(bins.shape)


## run script interactively

def runscript(pimax,boxsize,npoints):
    array_galaxy = numpy.random.uniform(0.,boxsize,npoints*3).reshape(-1,3)

    bins = 10.**numpy.linspace(numpy.log10(0.1),numpy.log10(30.),40)

    ## compute correlations
    DD = data_data(bins,pimax,array_galaxy,boxsize)
    RR = random_random(bins,pimax,npoints,boxsize)

    w_p = (DD[1:]/RR[1:] - 1.) * pimax * 2.

    # output
    print "projected correlation function:"
    print w_p
    print ""

    ## compute real-space correlations
    DD_xsi = data_data_3d(bins,array_galaxy,boxsize)
    RR_xsi = random_random_3d(bins,npoints,boxsize)

    xsi = (DD_xsi[1:]/RR_xsi[1:] - 1.)

    print "real-space correlation function:"
    print xsi
    print ""

    return bins,w_p,xsi

# load any generated galaxies from disk
data_data.load()
data_data_3d.load()

bins,wp,xsi = runscript(60., 180.7, 2e4)

# cache our generated galaxies on disk
data_data.dump()
data_data_3d.dump()

print data_data.info()
print data_data_3d.info()



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2016 California Institute of Technology.
# Copyright (c) 2016-2018 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/klepto/blob/master/LICENSE
"""
custom caching dict, which archives results to memory, file, or database
"""
from __future__ import absolute_import
from ._archives import cache
from ._archives import dict_archive as _dict_archive
from ._archives import null_archive as _null_archive
from ._archives import dir_archive as _dir_archive
from ._archives import file_archive as _file_archive
from ._archives import sql_archive as _sql_archive
from ._archives import sqltable_archive as _sqltable_archive
from ._archives import _sqlname

__all__ = ['cache','dict_archive','null_archive','dir_archive',\
           'file_archive','sql_archive','sqltable_archive']

class dict_archive(_dict_archive):
    def __new__(dict_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with an in-memory dictionary archive backend

    Inputs:
        name: (optional) identifier string [default: None]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        """
        if dict is None: dict = {}
        archive = _dict_archive()
        archive.__state__['id'] = str(name)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass

class null_archive(_null_archive):
    def __new__(null_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with a permanently-empty archive backend

    Inputs:
        name: (optional) identifier string [default: None]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        """
        if dict is None: dict = {}
        archive = _null_archive()
        archive.__state__['id'] = str(name)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass

class dir_archive(_dir_archive):
    def __new__(dir_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with a file-folder archive backend

    Inputs:
        name: name of the root archive directory [default: memo]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        serialized: if True, pickle file contents; otherwise save python objects
        compression: compression level (0 to 9) [default: 0 (no compression)]
        memmode: access mode for files, one of {None, 'r+', 'r', 'w+', 'c'}
        memsize: approximate size (in MB) of cache for in-memory compression
        """
        if dict is None: dict = {}
        archive = _dir_archive(name, **kwds)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass

class file_archive(_file_archive):
    def __new__(file_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with a single file archive backend

    Inputs:
        name: name of the file archive [default: memo.pkl or memo.py]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        serialized: if True, pickle file contents; otherwise save python objects
        """
        if dict is None: dict = {}
        archive = _file_archive(name, **kwds)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass

class sqltable_archive(_sqltable_archive):
    def __new__(sqltable_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with a sql database table archive backend

    Connect to an existing database, or initialize a new database, at the
    selected database url. For example, to use a sqlite database 'foo.db'
    in the current directory, database='sqlite:///foo.db'. To use a mysql
    database 'foo' on localhost, database='mysql://user:pass@localhost/foo'.
    For postgresql, use database='postgresql://user:pass@localhost/foo'. 
    When connecting to sqlite, the default database is ':memory:'; otherwise,
    the default database is 'defaultdb'. Connections should be given as
    database?table=tablename; for example, name='sqlite:///foo.db?table=bar'.
    If not provided, the default tablename is 'memo'. If sqlalchemy is not
    installed, storable values are limited to strings, integers, floats, and
    other basic objects. If sqlalchemy is installed, additional keyword
    options can provide database configuration, such as connection pooling.
    To use a mysql or postgresql database, sqlalchemy must be installed.

    Inputs:
        name: url for the sql database and table [default: (see note above)]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        serialized: if True, pickle table contents; otherwise cast as strings
        """
        if dict is None: dict = {}
        db, table = _sqlname(name)
        archive = _sqltable_archive(db, table, **kwds)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass

class sql_archive(_sql_archive):
    def __new__(sql_archive, name=None, dict=None, cached=True, **kwds):
        """initialize a dictionary with a sql database archive backend

    Connect to an existing database, or initialize a new database, at the
    selected database url. For example, to use a sqlite database 'foo.db'
    in the current directory, database='sqlite:///foo.db'. To use a mysql
    database 'foo' on localhost, database='mysql://user:pass@localhost/foo'.
    For postgresql, use database='postgresql://user:pass@localhost/foo'. 
    When connecting to sqlite, the default database is ':memory:'; otherwise,
    the default database is 'defaultdb'. If sqlalchemy is not installed,
    storable values are limited to strings, integers, floats, and other
    basic objects. If sqlalchemy is installed, additional keyword options
    can provide database configuration, such as connection pooling.
    To use a mysql or postgresql database, sqlalchemy must be installed.

    Inputs:
        name: url for the sql database [default: (see note above)]
        dict: initial dictionary to seed the archive
        cached: if True, use an in-memory cache interface to the archive
        serialized: if True, pickle table contents; otherwise cast as strings
        """
        if dict is None: dict = {}
        archive = _sql_archive(name, **kwds)
        if cached: archive = cache(archive=archive)
        archive.update(dict)
        return archive
    pass


# EOF



from otm_hvi_new import eureka as eureka_new
from otm_hvi import eureka as eureka_old
def new_model(h,v):
  return eureka_new([h,v,0.0])

def old_model(h,v):
  return eureka_old([h,v,0.0])

from numpy import vectorize
new_modelv = vectorize(new_model)
old_modelv = vectorize(old_model)

from numpy import mgrid    
h,v = mgrid[0.5:3.0:100j, 4.5:7.0:100j]
eo = old_modelv(h,v)
en = new_modelv(h,v)
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.axes import subplot_class_factory
Subplot3D = subplot_class_factory(Axes3D)

fig = plt.figure()
ax1 = Subplot3D(fig, 2,1,1)
ax1.plot_surface(h,v,eo, cmap=plt.cm.jet, rstride=5, cstride=5)
ax1.set_title('old_model')
ax1.set_xlabel('h')
ax1.set_ylabel('v')

ax2 = Subplot3D(fig, 2,1,2)
ax2.plot_surface(h,v,en, cmap=plt.cm.jet, rstride=5, cstride=5)
ax2.set_title('new model')
ax2.set_xlabel('h')
ax2.set_ylabel('v')
plt.show()




##############################
# [ h    v    a ] & [ Y ]
h_lower = 0.5;  h_upper = 3.0
v_lower = 4.5;  v_upper = 7.0
a_lower = 0.0;  a_upper = 0.0
Y_lower = 8.0;  Y_upper = 35.0
gridsize = 100
##############################

# settings
hausdorff = hdn = (1.0, 1.0, 1.0, 1.0)
Cy = .01
Cx = [Cy * xi for xi in hausdorff[:-1]]
tol = 0.0
verbose = False #True
settings = {'hdn':hdn, 'tol':tol, 'Cx':Cx, 'Cy':Cy, 'verbose':verbose}

# optimizer settings
maxiter = 200 #50
npop = 20
imax = 300 #30
ipop = 20
settings.update({'maxiter': maxiter, 'npop':npop, 'imax':imax, 'ipop':ipop})

from otm_hvi_new import eureka as model

# bounds
lower_bounds = (h_lower, v_lower, a_lower)
upper_bounds = (h_upper, v_upper, a_upper)
bounds = (lower_bounds, upper_bounds)
Y_bounds = (Y_lower, Y_upper)


# EOF



"""Tests for httplib2 when the socket module is missing.

This helps ensure compatibility with environments such as AppEngine.
"""
import os
import sys
import unittest

import httplib2

class MissingSocketTest(unittest.TestCase):
    def setUp(self):
        self._oldsocks = httplib2.socks
        httplib2.socks = None

    def tearDown(self):
        httplib2.socks = self._oldsocks

    def testProxyDisabled(self):
        proxy_info = httplib2.ProxyInfo('blah',
                                        'localhost', 0)
        client = httplib2.Http(proxy_info=proxy_info)
        self.assertRaises(httplib2.ProxiesUnavailableError,
                          client.request, 'http://localhost:-1/')



# h5py: store objects in a HDF5 hierarchy
'''
h5py allows for a single file to contain a hierarchy of stored objects

a h5py.File object has attrs, which has a (nearly?) complete dict API
however, it appears only basic objects can be stored

could convert objects to pickled strings, but doesn't allow embedded NULLs
so... how convert objects to strings?  (hickle does something like this)

could use h5py for dict in one file, and hickle for directory of files as dict

AHA...
storing any object as a dataset composed of a single pickled object WORKS

also, h5py.File object name and attrs can take protocol=0 strings as names
(thus, we *can* use the 'attrs' dict interface to store our objects)
'''
a = 1
b = 'hello'
d = {'x':3}
class e(object):
  pass
f = lambda x:x
g = e()

"""
import h5py
f = h5py.File('test.hkl','w')
f.attrs['a'] = a
print f.attrs['a']

f.attrs['b'] = b
print f.attrs['b']

f.close()
"""

# however, using hickle's trick... we can store any object as data
import h5py as h5
h = h5.File('test.hkl', 'w')
import dill      
d0 = h.create_dataset(dill.dumps(a,0), data=[dill.dumps(a)])
print dill.loads(d0[0])
d1 = h.create_dataset(dill.dumps(b,0), data=[dill.dumps(b)])
print dill.loads(d1[0])
d2 = h.create_dataset(dill.dumps(d,0), data=[dill.dumps(d)])
print dill.loads(d2[0])
d3 = h.create_dataset(dill.dumps(e,0), data=[dill.dumps(e)])
print dill.loads(d3[0])
d4 = h.create_dataset(dill.dumps(f,0), data=[dill.dumps(f)])
print dill.loads(d4[0])
d5 = h.create_dataset(dill.dumps(g,0), data=[dill.dumps(g)])
print dill.loads(d5[0])

#XXX: use metadata to determine if store directly or as pickled string?
#XXX: use metadata to store object type?
# d0.attrs['type'] = str(type(a))

h.attrs[dill.dumps(a,0)] = dill.dumps(a,0)
print dill.loads(h.attrs[dill.dumps(a,0)])
h.attrs[dill.dumps(b,0)] = dill.dumps(b,0)
print dill.loads(h.attrs[dill.dumps(b,0)])
h.attrs[dill.dumps(d,0)] = dill.dumps(d,0)
print dill.loads(h.attrs[dill.dumps(d,0)])
h.attrs[dill.dumps(e,0)] = dill.dumps(e,0)
print dill.loads(h.attrs[dill.dumps(e,0)])
h.attrs[dill.dumps(f,0)] = dill.dumps(f,0)
print dill.loads(h.attrs[dill.dumps(f,0)])
h.attrs[dill.dumps(g,0)] = dill.dumps(g,0)
print dill.loads(h.attrs[dill.dumps(g,0)])

h.close()





import os
import unittest

import httplib2

from httplib2.test import miniserver


class HttpSmokeTest(unittest.TestCase):
    def setUp(self):
        self.httpd, self.port = miniserver.start_server(
            miniserver.ThisDirHandler)

    def tearDown(self):
        self.httpd.shutdown()

    def testGetFile(self):
        client = httplib2.Http()
        src = 'miniserver.py'
        response, body = client.request('http://localhost:%d/%s' %
                                        (self.port, src))
        self.assertEqual(response.status, 200)
        self.assertEqual(body, open(os.path.join(miniserver.HERE, src)).read())



#!/usr/bin/env python
"""
OUQ with model sausage: valid_wrt_model, mean(y), norm(wts)
  with 'wiggle room' around the model in Y
"""

hausdorff = (1.0, 1.0, 1.0, 1.0) # spread(x) where x = h,v,a,Y
cde = 20 #NOTE: int is DE(cde), 0 is powell(); used in graphical_distance
ide = 10 #NOTE: int is DE(ide), 0 is powell(); used in graphical_distance
debug = False
MINMAX = -1  ## NOTE: sup = maximize = -1; inf = minimize = 1
from mystic.math.measures import mean
#######################################################################
# optimizer configuration parameters
#######################################################################
npop = 20 #16 #32 # 40 #50  #!!!
maxiter = 1000
maxfun = 1e+6
convergence_tol = 1e-6; ngen = 40 #20  #!!!
smearing = 0.001
crossover = 0.9
percent_change = 0.9

# MPI
nodes = '25:ppn=2' # using mp_map in solver, means 4 processor per (Cy,theta)
queue = 'weekdayQ'
timelimit = '11:59:59'


#######################################################################
# the model function
#######################################################################
#from otm_hvi import eureka as model
from otm_hvi_new import eureka as model


#######################################################################
# the differential evolution optimizer
#######################################################################
def optimize(cost,_bounds,_constraints,initial=None,job='',stepmon=None):
  from mystic.solvers import DifferentialEvolutionSolver2
  from mystic.termination import ChangeOverGeneration as COG
  from mystic.strategy import Best1Exp
  from mystic.monitors import Null #, Monitor
 #from mystic.tools import random_seed
 #random_seed(123)

  if stepmon is None:
    logfile = 'paramlog'
    stamp = "_".join([logfile, str(job)])
    from mystic.monitors import VerboseLoggingMonitor #, LoggingMonitor
    stepmon = VerboseLoggingMonitor(1,10,filename=stamp+'.txt',info=stamp)
# evalmon = Monitor()
  evalmon = Null()

  lb,ub = _bounds
  ndim = len(lb)

  solver = DifferentialEvolutionSolver2(ndim,npop)
  if initial:
    solver.SetInitialPoints(initial,radius=smearing)
  else:
    solver.SetRandomInitialPoints(min=lb,max=ub)
 #ALTERNATE: initial population
 #solver.SetRandomInitialPoints(min=lb,max=ub)
 #if initial:
 #  solver.population[0] = initial 
  solver.SetStrictRanges(min=lb,max=ub)
  solver.SetEvaluationLimits(maxiter,maxfun)
  solver.SetEvaluationMonitor(evalmon)
  solver.SetGenerationMonitor(stepmon)
  solver.SetConstraints(_constraints)
 #ALTERNATE: multiprocessing map of cost function
##from pathos.mp_map import mp_map as map  # multiprocessing
##solver.SetMapper(map)
 #ALTERNATE: MPI map of cost function
##from pyina.ez_map import ez_map2 as map  # mpi4py
##from pyina.launchers import mpirun_launcher, torque_launcher
##from pyina.mappers import carddealer_mapper, equalportion_mapper
##from pyina.schedulers import torque_scheduler
##solver.SetMapper(map, equalportion_mapper)
##solver.SetLauncher(mpirun_launcher, nodes)
##solver.SelectScheduler(torque_scheduler, queue, timelimit)
  tol = convergence_tol
  solver.Solve(cost,termination=COG(tol,ngen),strategy=Best1Exp, \
               CrossProbability=crossover,ScalingFactor=percent_change)

  solved = solver.bestSolution
 #print "solved: %s" % solver.Solution()
  func_max = MINMAX * solver.bestEnergy       #NOTE: -solution assumes -Max
 #func_max = 1.0 + MINMAX*solver.bestEnergy   #NOTE: 1-sol => 1-success = fail
  func_evals = solver.evaluations
 #from mystic.munge import write_support_file
 #write_support_file(stepmon, log_file=stamp+'.py', header=stamp)#, npts=npts)
  return solved, func_max, func_evals


#######################################################################
# maximize the function
#######################################################################
def maximize(params,npts,bounds,initial=None,job='',printmon=None):

  from my_dirac_measure import scenario
  from numpy import inf
  target,error = params
  lb,ub = bounds

  # NOTE: rv, lb, ub are of the form:
  #    rv = [wxi]*nx + [xi]*nx + [wyi]*ny + [yi]*ny + [wzi]*nz + [zi]*nz
  #       +  [Yi]*nY

  # generate secondary constraints function
  def more_constraints(rv):
    ##################### begin function-specific #####################
    ###################### end function-specific ######################
    return rv

# from memoize import memoized0_nopickle_round as memoized

  # generate primary constraints function
# @memoized(tol=3)
  def constraints(rv):
    rrv = range(len(rv))
    for i in rrv:
      rv[i] = min(max(rv[i], lb[i]), ub[i])
    # impose range constraints: first try bounce-back to maintain randomness
#   rv = [2*lb[i]-rv[i] if max(rv[i], lb[i]) == lb[i] else rv[i] for i in rrv]
#   rv = [2*ub[i]-rv[i] if min(rv[i], ub[i]) == ub[i] else rv[i] for i in rrv]
    # if bounceback fails, then use a hard cutoff  #FIXME: breaks randomness
#   rv = [max(rv[i], lb[i]) for i in rrv]
#   rv = [min(rv[i], ub[i]) for i in rrv]

    c = scenario()
    c.load(rv, npts)

    # ensure norm(wi) = 1.0 in each discrete measure
    norm = 1.0
    for i in range(len(c)):
      w = c[i].weights
      w[-1] = norm - sum(w[:-1])
      c[i].weights = w

    ##################### begin function-specific #####################
    # impose mean on the values of the product measure
    from mystic.math.dirac_measure import mean_y_norm_wts_constraintsFactory as factory
    constrain = factory((target[0],error[0]), npts)

    # check mean value, and if necessary use constrain to set mean value
    y = float(mean(c.values, c.weights))
    if not (y >= float(target[0] - error[0])):
      c.update( constrain( c.flatten(all=True) ) )

    # then test if valid... then impose model validity on product measure
    if not c.valid_wrt_model(model, ytol=target[2], xtol=target[3], \
                             imax=target[4], ipop=cde, hausdorff=hausdorff):
      c.set_valid(model, cutoff=target[2], bounds=bounds, tol=error[2], \
                  constraints=constrain, xtol=target[3], maxiter=error[3], \
                  imax=error[4], ipop=ide, hausdorff=hausdorff)
    ###################### more function-specific #####################
    if debug:
      if not c.valid_wrt_model(model, ytol=target[2], xtol=target[3], \
                               imax=target[4], ipop=cde, hausdorff=hausdorff):
        print "valid_wrt_model: False"
      if not [sum(w) for w in c.wts] == [1.0] * len(c.wts):
        print "norm(w) == 1.0: False"
      if not c.get_mean_value() >= float(target[0] - error[0]):
        print "mean(y) >= %s: False" % str(target[0] - error[0])
    ###################### end function-specific ######################
    # extract weights and positions and values
    return c.flatten(all=True)

  # generate a function that returns True for 'success' and False for 'failure'
  def safe(f):
    y = f - target[1]  #XXX: thus... if [f - theta] <= 0: then 'failure'
    return 0.0 if y <= 0.0 else y

  # generate maximizing function
  def cost(rv):
    c = scenario()
    c.load(rv, npts)
    #XXX: apply 'filters' to catch errors in constraints solver (necessary ?)
    ##################### begin function-specific #####################
    y = float(mean(c.values, c.weights))
    if not (y >= float(target[0] - error[0])):
      if debug: print "skipping mean: %s" % y
      return inf  #XXX: FORCE TO SATISFY E CONSTRAINTS

    if not c.valid_wrt_model(model, ytol=target[2], xtol=target[3], \
                             imax=target[4], ipop=cde, hausdorff=hausdorff):
      if debug: print "skipping model-invalidity"
      return inf  #XXX: FORCE TO SATISFY E CONSTRAINTS
    ###################### end function-specific ######################
    return MINMAX * c.pof_value(safe)

  # maximize
  solved, func_max, func_evals = optimize(cost,(lb,ub),constraints, \
                                          initial,job,printmon)

  # clear the cache
# constraints.memo.clear()

  if MINMAX == 1:
    tag = 'minimum' # inf
  else:
    tag = 'maximum' # sup
  if printmon:
    printmon.info("func_%s: %s" % (tag,func_max))
    printmon.info("func_evals: %s" % func_evals)
  else:
    print "func_%s: %s" % (tag,func_max)
    print "func_evals: %s" % func_evals

  return solved, func_max


def outer_loop(initial=None, **settings):
  """
  initial: initial optimization parameter values, often taken from prior runs
  settings: a dict of settings (such as theta, Cy, etc) to override

  Returns these same parameters.  If convergence is unsatisfactory, should
  trigger a rerun of current settings instead of moving on.
  """
  function_name = model.__name__

  y_mean = 12.0   #NOTE: SET THE 'mean' HERE!
  y_mean_error = 0.0 #NOTE: SET THE 'error' HERE!
  theta = 8.0      #NOTE: SET THE 'failure criteria' HERE!
  Cy = 3.0           #NOTE: SET THE 'cutoff' HERE!
  Cx = (0.0,0.0,0.0,0.0) #NOTE: SET THE 'wiggle' HERE!
  valid_tol = 0.0  #NOTE: SET THE 'model tolerance' HERE!
  imax_c = 200     #NOTE: SET THE 'max iterations to check valid' HERE!
  imax_i = 10      #NOTE: SET THE 'max inner iterations to set valid' HERE!
  imax_o = 50      #NOTE: SET THE 'max outer iterations to set valid' HERE!

  #APPLY THE SETTINGS OVERRIDES HERE#
  for k,v in settings.items():
    exec("%s = %s" % (str(k),repr(v))) #XXX: HACK

  nx = 2  #NOTE: SET THE NUMBER OF 'h' POINTS HERE!
  ny = 2  #NOTE: SET THE NUMBER OF 'v' POINTS HERE!
  nz = 1  #NOTE: SET THE NUMBER OF 'a' POINTS HERE!

  #XXX: check units versus those in model
  w_lower = [0.0]; Y_lower = [theta]
  w_upper = [1.0]; Y_upper = [30.0+Cy]#[100.0]
  h_lower = [0.5]; a_lower = [0.0];  v_lower = [4.5]
  h_upper = [3.0]; a_upper = [0.0];  v_upper = [7.0]

  target = (y_mean,theta,Cy,Cx,imax_c)
  error = (y_mean_error,None,valid_tol,imax_o,imax_i)
  pars = (target,error)

  npts = (nx,ny,nz)
  from mystic.math.paramtrans import _npts
  _n = _npts(npts)

 #XXX: EDITED TO FIX Y1 == Y_lower
  lower_bounds = (nx * w_lower) + (nx * h_lower) \
               + (ny * w_lower) + (ny * v_lower) \
               + (nz * w_lower) + (nz * a_lower) \
               + (_n * Y_lower)
  upper_bounds = (nx * w_upper) + (nx * h_upper) \
               + (ny * w_upper) + (ny * v_upper) \
               + (nz * w_upper) + (nz * a_upper) \
               + (Y_lower) + (Y_upper)*(_n - 1)
#              + (Y_upper) + (Y_lower)*(_n - 1) #XXX: probably true
  bounds = (lower_bounds,upper_bounds)

  import datetime
  infofile = 'paramlog'
  _stamp = "_".join([infofile, str(job)])
  from mystic.monitors import VerboseLoggingMonitor
  printmon = VerboseLoggingMonitor(1,1,filename=_stamp+'.txt',info=_stamp)
  printmon.info("%s" % datetime.datetime.now().ctime())
# from mystic.monitors import LoggingMonitor
# printmon = LoggingMonitor(1,filename=_stamp+'.txt',info=_stamp)
  printmon.info("...SETTINGS...")
  printmon.info("npop = %s" % npop)
  printmon.info("maxiter = %s" % maxiter)
  printmon.info("maxfun = %s" % maxfun)
  printmon.info("convergence_tol = %s" % convergence_tol)
  printmon.info("crossover = %s" % crossover)
  printmon.info("percent_change = %s" % percent_change)
  printmon.info("..............\n")

  printmon.info(" model: f(x) = %s(x)" % function_name)
  printmon.info(" target: %s" % str(target))
  printmon.info(" error: %s" % str(error))
  printmon.info(" npts: %s" % str((nx,ny,nz)))
  printmon.info("..............\n")

  param_string = "["
  for i in range(nx):
    param_string += "'wx%s', " % str(i+1)
  for i in range(nx):
    param_string += "'x%s', " % str(i+1)
  for i in range(ny):
    param_string += "'wy%s', " % str(i+1)
  for i in range(ny):
    param_string += "'y%s', " % str(i+1)
  for i in range(nz):
    param_string += "'wz%s', " % str(i+1)
  for i in range(nz):
    param_string += "'z%s', " % str(i+1)
  for i in range(_n):
    param_string += "'Y%s', " % str(i+1)
  param_string = param_string[:-2] + "]"

  printmon.info(" parameters: %s" % param_string)
  printmon.info(" lower bounds: %s" % lower_bounds)
  printmon.info(" upper bounds: %s" % upper_bounds)
# print " ..."

 #pars = (target,error)
 #npts = (nx,ny,nz)
 #bounds = (lower_bounds,upper_bounds)
  solved, diameter = maximize(pars,npts,bounds,initial,job,printmon)

  from numpy import array
  from my_dirac_measure import scenario
  c = scenario()
  c.load(solved,npts)
  printmon.info("solved[wx,x]: %s" % zip(c[0].weights,c[0].coords))
  printmon.info("solved[wy,y]: %s" % zip(c[1].weights,c[1].coords))
  printmon.info("solved[wz,z]: %s" % zip(c[2].weights,c[2].coords))
  printmon.info("solved[Y]: %s" % c.values)
 #print "solved: %s" % str( c.flatten(all=True) )

  printmon.info("fails valid wrt model: %s" % \
        c.valid_wrt_model(model, xtol=Cx, ytol=Cy, blamelist=True, \
                          imax=imax_c, ipop=cde, hausdorff=hausdorff))
  printmon.info("mean(y): %s >= %s" % (str( c.get_mean_value() ), y_mean - y_mean_error))
  printmon.info("sum_wts: %s == 1.0" % [sum(w) for w in c.wts])

# try: _Cx = max(Cx)
# except TypeError: _Cx = Cx
  from mystic.math.paramtrans import graphical_distance
  Ry = graphical_distance(model, c, ytol=Cy, xtol=Cx, cutoff=0.0, \
                          imax=0, ipop=cde, hausdorff=hausdorff)
  printmon.info("vertical_distance: %s" % Ry)#, Cy + _Cx))
  Rv = graphical_distance(model, c, ytol=Cy, xtol=Cx, cutoff=0.0, \
                          imax=imax_c, ipop=cde, hausdorff=hausdorff)
  printmon.info("graphical_distance: %s <= %s" % (Rv, Cy))
  printmon.info("%s" % datetime.datetime.now().ctime())

  if False: #XXX: return 'initial' instead of 'solved' when solver fails ?
    return initial, settings#, False
  else:
    return solved, settings#, True


#######################################################################
# rank, bounds, and restart information 
#######################################################################
if __name__ == '__main__':

  #XXX XXX XXX ADJUST THIS CODE AS NEEDED TO PERFORM ALL RUNS XXX XXX XXX
  def looper(variables): #XXX: HACK
    job, (initial,Cy,theta) = variables
    job = "_".join([str(Cy),str(theta),str(job)])#.replace('.','')
    Cx = [Cy, Cy, 0.0]
   #return outer_loop(initial,Cy=Cy,Cx=0.0,theta=theta,job=job)
    return outer_loop(initial,Cy=Cy,Cx=Cx,theta=theta,job=job)

  Cys = [1.0]
  thetas = [11.0, 10.0, 9.0, 8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0, 1.0, 0.0] 
  initial = None
  #XXX XXX XXX ---------------------------------------------- XXX XXX XXX
  Cys = [0.12]   #XXX: TEST
  thetas = [7.0] #XXX: TEST

  # get all the possible runs
  from mystic.math.measures import _pack
  runs = _pack([Cys, thetas])
  runs = list(enumerate([(initial,_cy,_theta) for (_cy,_theta) in runs]))
  print "queue = %s" % dict(runs)

  # with 'serial' python
  results = map(looper, runs)
  # with multiprocessing 
  ##nodes = 3 #len(runs)
  ##from pathos.mp_map import mp_map as map  # multiprocessing
  ##results = map(looper, runs, nproc=nodes)


# EOF



"""
a decorator named 'memoized' that can memoize a *method* on an object.
"""
def isiterable(x):
 #try:
 #    from collections import Iterable
 #    return isinstance(x, Iterable)
 #except ImportError:
  try:
    iter(x)
    return True
  except TypeError: return False

def deep_round_factory(tol):
  def deep_round(*args, **kwds):
    argstype = type(args) 
    _args = list(args)
    _kwds = kwds.copy()
    for i,j in enumerate(args):
      if isinstance(j, float): _args[i] = round(j, tol) # don't round int
      elif isinstance(j, (str, unicode, type(BaseException()))): continue
      elif isinstance(j, dict): _args[i] = deep_round(**j)[1]
      elif isiterable(j): #XXX: fails on the above, so don't iterate them
        jtype = type(j)
        _args[i] = jtype(deep_round(*j)[0])
    for i,j in kwds.items():
      if isinstance(j, float): _kwds[i] = round(j, tol)
      elif isinstance(j, (str, unicode, type(BaseException()))): continue
      elif isinstance(j, dict): _kwds[i] = deep_round(**j)[1]
      elif isiterable(j): #XXX: fails on the above, so don't iterate them
        jtype = type(j)
        _kwds[i] = jtype(deep_round(*j)[0])
    return argstype(_args), _kwds
  return deep_round

"""
>>> deep_round = deep_round_factory(tol=0)  #FIXME: convert to decorator !!!
>>> deep_round([1.12,2,{'x':1.23, 'y':[4.56,5.67]}], x=set([11.22,44,'hi']))
(([1.0, 2, {'y': [5.0, 6.0], 'x': 1.0}],), {'x': set([11.0, 'hi', 44])})
"""

def deep_round(tol=0):
  def dec(f):
    def func(*args, **kwds):
      _deep_round = deep_round_factory(tol)
      _args,_kwds = _deep_round(*args, **kwds)
      return f(*_args, **_kwds)
    return func
  return dec


#FIXME: the below need expiration of cache due to time, calls, etc...
#       and potentially r/w to database, file, or other caching mechanism
#FIXME: memoize*_round fails when decorating a class method


def memoized0_nopickle_round(tol=0):
    """Decorator that memoizes a function's return value each time it is called.
    If called later with the same arguments, the memoized value is returned, and
    not re-evaluated.  This may lead to memory issues, as memo is never cleared.
    This decorator takes an integer tolerance 'tol', equal to the number of
    decimal places to which it will round off floats.
    This funciton does not pickle, and thus can only handle basic types.
    """
    memo = {}

    @deep_round(tol)
    def rounded_args(*args, **kwds):
        return (args, kwds)

    def dec(f):
        def func(*args, **kwds):
            try:
                _args, _kwds = rounded_args(*args, **kwds)
                argstr = str((_args, _kwds))
                if not memo.has_key(argstr):
                    memo[argstr] = f(*args, **kwds)
                return memo[argstr]
            except: #TypeError
                return f(*args, **kwds)
        func.memo = memo
        return func
    return dec


def memoized0_round(tol=0):
    """Decorator that memoizes a function's return value each time it is called.
    If called later with the same arguments, the memoized value is returned, and
    not re-evaluated.  This may lead to memory issues, as memo is never cleared.
    This decorator takes an integer tolerance 'tol', equal to the number of
    decimal places to which it will round off floats.
    """
    memo = {}

    @deep_round(tol)
    def rounded_args(*args, **kwds):
        return (args, kwds)

    def dec(f):
        def func(*args, **kwds):
            try:
                _args, _kwds = rounded_args(*args, **kwds)
               #import cPickle as pickle
                import dill as pickle
                argstr = pickle.dumps((_args, _kwds))
                if not memo.has_key(argstr):
                    memo[argstr] = f(*args, **kwds)
                return memo[argstr]
            except: #TypeError
                return f(*args, **kwds)
        func.memo = memo
        return func
    return dec


def memoized0(f):
    """Decorator that memoizes a function's return value each time it is called.
    If called later with the same arguments, the memoized value is returned, and
    not re-evaluated.  This may lead to memory issues, as memo is never cleared.
    """
    memo = {}

    def func(*args, **kwds):
        try:
           #import cPickle as pickle
            import dill as pickle
            argstr = pickle.dumps((args, kwds))
            if not memo.has_key(argstr):
                memo[argstr] = f(*args, **kwds)
            return memo[argstr]
        except: #TypeError
            return f(*args, **kwds)
    func.memo = memo
    return func


class memoized(object):
    """Decorator that memoizes a function's return value each time it is called.
    If called later with the same arguments, the memoized value is returned, and
    not re-evaluated.  This may lead to memory issues, as memo is never cleared.
    """
    def __init__(self, func):
      self.func = func
      self.memo = {}

    def __call__(self, *args, **kwds):
      try:
         #import cPickle as pickle
          import cPickle as pickle
          argstr = pickle.dumps((args, kwds))
          if not self.memo.has_key(argstr):
              self.memo[argstr] = self.func(*args, **kwds)
          return self.memo[argstr]
      except: #TypeError
          return self.func(*args, **kwds)

    def __repr__(self):
      """Return the function's docstring."""
      return self.func.__doc__

    def __get__(self, obj, objtype):
      """Support instance methods."""
      import functools
      return functools.partial(self.__call__, obj)

# EOF



#! /usr/bin/env python
"""
...associated transformations for legacy data module

lipschitz_distance: ...

graphical_distance: for a given dataset (x,y) and a given model (F),
  find the radius(x') that minimizes the graph between reality, y = G(x),
  and an approximating function, y' = F(x')
"""
debug = False 

def absolute_distance(x, xp, up=False, dmin=0):
  """distance = |x - x'|;  (see euclidean_distance for notes)"""
  from numpy import abs, asarray, newaxis as nwxs
  from __builtin__ import max
  # cast as arrays of the same dimension
  x = asarray(x); xp = asarray(xp)
  xsize = max(len(x.shape), len(xp.shape), dmin)
  while len(x.shape) < xsize: x = x[nwxs]
  while len(xp.shape) < xsize: xp = xp[nwxs]
  # 'upconvert' with a newaxis  #XXX: only works correctly for dim=0,1,2 ?
  if up: x = x[:,nwxs] if xsize > 0 else x[nwxs]
  return abs(x - xp)

def euclidean_distance(x, xp, up=True, dmin=0):
  """1-D euclidean distance between points

  d[i] = | x[i] - x'[i] |

Input:
  up   = True if upconvert x with x[:,newaxis]
  dmin = upconvert to x,x' to dimension >= dmin

Notes:
  for standard array behavior, use up=False
  for element-wise across all elements, use up=True
"""
  return absolute_distance(x,xp,up=up,dmin=dmin).swapaxes(0,1).T

def manhattan_distance(x, xp, **kwds):
  """1-D manhattan distance between points

  d[ij] = | x[i] - x'[j] |

Notes:
  manhattan matrix where x'==x will be symmetric with zeros on the diagonal 
  use dmin=2 for forced upconversion of 1-D arrays
"""
  # dmin: force upconvert to x,x' to dimension >= dmin
  dmin = 0 # default
  if kwds.has_key('dmin'):
    dmin = kwds['dmin']
  from numpy import abs, asarray, newaxis as nwxs
  from __builtin__ import max
  # cast as arrays of the same dimension
  x = asarray(x); xp = asarray(xp)
  xsize = max(len(x.shape), len(xp.shape), dmin)
  while len(x.shape) < xsize: x = x[nwxs]
  while len(xp.shape) < xsize: xp = xp[nwxs]
  # prep for build manhattan matrix in single operation
  if xsize >= 2:
    x = x.T[:,:,nwxs]; xp = xp.T[:,nwxs]
  elif xsize == 1:
    x = x[:,nwxs]#     x.T[:,nwxs];  xp = xp.T
  else:#xsize == 0:
    x = x[nwxs]
  return abs(x - xp)#.T

###############################################################################
# Note: manhattan and euclidean are now essentially 'the same'...
#       (is sum done correctly? is orientation correct? ...)
###############################################################################

def lipschitz_metric(L, x, xp):
  """sum of lipschitz-weighted distance between points

  d = sum( L[i] * |x[i] - x'[i]| )
"""
  #FIXME: merge with lipschitz cone (distance/contains)
  from numpy import sum, asarray
  d = manhattan_distance(x,xp) #XXX: ,dmin=2)
  return sum(L * d.T, axis=-1).T

def _npts(*x):
  """get len(product measure), given lengths of each underlying measure"""
  from numpy import product
  return product(x)

def _get_xy(points):
  """extract the list of positions and the list of values for given points"""
  from numpy import asarray
  if hasattr(points, 'coords'):
    x  = points.coords
    y  = points.values
  elif len(asarray(points).shape) >= 1:
    x = [p.position for p in points] 
    y = [p.value for p in points] 
  else:
    x = points.position
    y = points.value
  return x,y

###########################################################################
# distance metrics  #NOTE: euclidean_distance and the like need renaming!!!
###########################################################################

def chebyshev(x,xp, up=True, dmin=0, axis=None):
  """infinity norm distance between points in euclidean space

  d(inf) =  max( |x[0] - x[0]'|, |x[1] - x[1]'|, ..., |x[n] - x[n]'| ) 

Input:
  x    = array of points, x
  xp   = array pf points, x'

Additional Input:
  up   = True if upconvert x with x[:,newaxis]
  dmin = upconvert to x,x' to dimension >= dmin
  axis = if not None, reduce across the selected axis

Notes:
  for standard array behavior, use up=False
  for element-wise across all elements, use up=True
"""
  d = euclidean_distance(x,xp,up=up,dmin=dmin)
  from numpy import max
  return max(d, axis=axis)


def minkowski(x,xp, up=True, dmin=0, p=3, axis=None):
  """p-norm distance between points in euclidean space

  d(p) = sum( |x[0] - x[0]'|^p, |x[1] - x[1]'|^p, ..., |x[n] - x[n]'|^p )^(1/p)

Input:
  x    = array of points, x
  xp   = array pf points, x'

Additional Input:
  up   = True if upconvert x with x[:,newaxis]
  dmin = upconvert to x,x' to dimension >= dmin
  axis = if not None, reduce across the selected axis
  p    = value of p for the p-norm (default = 3)

Notes:
  for standard array behavior, use up=False
  for element-wise across all elements, use up=True
"""
  d = euclidean_distance(x,xp,up=up,dmin=dmin)
  from numpy import sum
  return sum(d**p, axis=axis)**(1./p)


def euclidean(x,xp, up=True, dmin=0, axis=None):
  """L-2 norm distance between points in euclidean space

  d(2) = sqrt(sum( |x[0] - x[0]'|^2, |x[1] - x[1]'|^2, ..., |x[n] - x[n]'|^2 ))

Input:
  x    = array of points, x
  xp   = array pf points, x'

Additional Input:
  up   = True if upconvert x with x[:,newaxis]
  dmin = upconvert to x,x' to dimension >= dmin
  axis = if not None, reduce across the selected axis

Notes:
  for standard array behavior, use up=False
  for element-wise across all elements, use up=True
"""
  return minkowski(x,xp,up=up,dmin=dmin,p=2,axis=axis)


def manhattan(x,xp, up=True, dmin=0, axis=None):
  """L-1 norm distance between points in euclidean space

  d(1) = sum( |x[0] - x[0]'|, |x[1] - x[1]'|, ..., |x[n] - x[n]'| )

Input:
  x    = array of points, x
  xp   = array pf points, x'

Additional Input:
  up   = True if upconvert x with x[:,newaxis]
  dmin = upconvert to x,x' to dimension >= dmin
  axis = if not None, reduce across the selected axis

Notes:
  for standard array behavior, use up=False
  for element-wise across all elements, use up=True
"""
  return minkowski(x,xp,up=up,dmin=dmin,p=1,axis=axis)

###########################################################################

def is_feasible(distance, cutoff=0.0):
  """determine if the distance exceeds the given cutoff distance

Input:
  distance = measure of feasibility for each point
  cutoff = maximum acceptable distance

Returns:
  True where the distance is less than cutoff
"""
  from numpy import asarray
  d = infeasibility(distance, cutoff) > 0.0
  if len(asarray(d).shape) is 0:
    return not d
  return -d


def infeasibility(distance, cutoff=0.0): 
  """amount by which the distance exceeds the given cutoff distance

Input:
  distance = measure of feasibility for each point
  cutoff = maximum acceptable distance

Returns:
  array of distances by which each point is infeasbile
"""
  from numpy import array
  distance = array(distance)
  # zero-out all distances less than tolerated
  if cutoff is not None:
    if len(distance.shape) is 0:
      return 0.0 if distance <= cutoff else distance
    distance[distance <= cutoff] = 0.0
  return distance


def lipschitz_distance(L, points1, points2, **kwds):
  """calculate the lipschitz distance between two sets of datapoints

Inputs:
 L = list of lipschitz constants
 points1 = dataset or list of 'datapoint' or 'cone.vertex' objects
 points2 = dataset or list of 'datapoint' or 'cone.vertex' objects

Additional Inputs:
 tol = maximum acceptable deviation from shortness
 cutoff = zero out distances less than cutoff; typically: tol, 0.0, or None

Returns:
 list of lipschitz distances

Notes:
 Each point x,y can be thought to have an associated double-cone with slope
 equal to the lipschitz constant. Shortness with respect to another point is
 defined by the first point not being inside the cone of the second. We can
 allow for some error in shortness, a short tolerance 'tol', for which the
 point x,y is some acceptable y-distance inside the cone. While very tightly
 related, cutoff and tol play distinct roles; tol is subtracted from
 calculation of the lipschitz_distance, while cutoff zeros out the value
 of any element less than the cutoff.
"""
  #FIXME: merge with lipschitz cone (distance/contains)
  x,y   = _get_xy(points1)
  xp,yp = _get_xy(points2)

  # get tolerance in y
  tol = kwds.pop('tol', 0.0)
  cutoff = tol  # default is to zero out distances less than tolerance
  if kwds.has_key('cutoff'): cutoff = kwds.pop('cutoff')
  if cutoff is True: cutoff = tol
  elif cutoff is False: cutoff = None

  # calculate the distance matrix
  md = manhattan_distance(y,yp) - max(0.0, tol)
  lm = lipschitz_metric(L,x,xp)
  d = md - lm
  # zero-out all distances less than tolerated 
  return infeasibility(d, cutoff)


def graphical_distance(model, points, **kwds):
  """find the radius(x') that minimizes the graph between reality, y = G(x),
and an approximating function, y' = F(x')

Inputs:
  model = the model function, y' = F(x'), that approximates reality, y = G(x)
  points = object of type 'datapoint' to validate against; defines y = G(x)

Additional Inputs:
  ytol = maximum acceptable difference |y - F(x')|; a single value
  xtol = maximum acceptable difference |x - x'|; an iterable or single value
  cutoff = zero out distances less than cutoff; typically: ytol, 0.0, or None
  hausdorff = norm; where if given, ytol = |y - F(x')| + |x - x'|/norm

Returns:
  radius = minimum distance from x,G(x) to x',F(x') for each x

Notes:
  xtol defines the n-dimensional base of a pilar of height ytol, centered at
  each point. The region inside the pilar defines the space where a "valid"
  model must intersect. If xtol is not specified, then the base of the pilar
  will be a dirac at x' = x. This function performs an optimization for each
  x to find an appropriate x'. While cutoff and ytol are very tightly related,
  they play a distinct role; ytol is used to set the optimization termination
  for an acceptable |y - F(x')|, while cutoff is applied post-optimization.
  If we are using the hausdorff norm, then ytol will set the optimization
  termination for an acceptable |y - F(x')| + |x - x'|/norm, where the x
  values are normalized by norm = hausdorff.
""" #FIXME: update docs to show normalization in y
 #NotImplemented:
 #L = list of lipschitz constants, for use when lipschitz metric is desired
 #constraints = constraints function for finding minimum distance
  from mystic.math.legacydata import dataset
  from numpy import asarray, sum, isfinite, zeros, seterr
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor

  # ensure target xe and ye is a dataset
  target = dataset()
  target.load(*_get_xy(points))
  nyi = target.npts             # y's are target.values
  nxi = len(target.coords[-1])  # nxi = len(x) / len(y)
  
  # NOTE: the constraints function is a function over a single xe,ye
  #       because each underlying optimization is over a single xe,ye.
  #       thus, we 'pass' on using constraints at this time...
  constraints = None   # default is no constraints
  if kwds.has_key('constraints'): constraints = kwds.pop('constraints')
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  # get tolerance in y and wiggle room in x
  ytol = kwds.pop('ytol', 0.0)
  xtol = kwds.pop('xtol', 0.0) # default is to not allow 'wiggle room' in x 

  cutoff = ytol  # default is to zero out distances less than tolerance
  if kwds.has_key('cutoff'): cutoff = kwds.pop('cutoff')
  if cutoff is True: cutoff = ytol
  elif cutoff is False: cutoff = None
  ipop = kwds.pop('ipop', min(20, 3*nxi)) #XXX: tune ipop?
  imax = kwds.pop('imax', 1000) #XXX: tune imax?

  # get range for the dataset (normalization for hausdorff distance)
  hausdorff = kwds.pop('hausdorff', False)
  if not hausdorff:  # False, (), None, ...
    ptp = [0.0]*nxi
    yptp = 1.0
  elif hausdorff is True:
    from mystic.math.measures import spread
    ptp = [spread(xi) for xi in zip(*target.coords)]
    yptp = spread(target.values)
  else:
    try: #iterables
      if len(hausdorff) < nxi+1:
        hausdorff = list(hausdorff) + [0.0]*(nxi - len(hausdorff)) + [1.0]
      ptp = hausdorff[:-1]  # all the x
      yptp = hausdorff[-1]  # just the y
    except TypeError: #non-iterables
      ptp = [hausdorff]*nxi
      yptp = hausdorff

  #########################################################################
  def radius(model, point, ytol=0.0, xtol=0.0, ipop=None, imax=None):
    """graphical distance between a single point x,y and a model F(x')"""
    # given a single point x,y: find the radius = |y - F(x')| + delta
    # radius is just a minimization over x' of |y - F(x')| + delta
    # where we apply a constraints function (of box constraints) of
    # |x - x'| <= xtol  (for each i in x)
    #
    # if hausdorff = some iterable, delta = |x - x'|/hausdorff
    # if hausdorff = True, delta = |x - x'|/spread(x); using the dataset range
    # if hausdorff = False, delta = 0.0
    #
    # if ipop, then DE else Powell; ytol is used in VTR(ytol)
    # and will terminate when cost <= ytol
    x,y = _get_xy(point)
    y = asarray(y)

    # build the cost function
    if hausdorff: # distance in all directions
      def cost(rv):
        '''cost = |y - F(x')| + |x - x'| for each x,y (point in dataset)'''
        errs = seterr(invalid='ignore', divide='ignore') # turn off warning 
        z = abs((asarray(x) - rv)/ptp)  # normalize by range
        m = abs(y - model(rv))/yptp     # normalize by range
        m = m if isfinite(m) else 10.0**6
        seterr(invalid=errs['invalid'], divide=errs['divide']) # turn on warning
        return m + sum(z[isfinite(z)])
    else:  # vertical distance only
      def cost(rv):
        '''cost = |y - F(x')| for each x,y (point in dataset)'''
        return abs(y - model(rv))

    if debug:
      print "rv: %s" % str(x)
      print "cost: %s" % cost(x)

    # if xtol=0, radius is difference in x,y and x,F(x); skip the optimization
    try:
      if not imax or not max(xtol): #iterables
        return cost(x)
    except TypeError:
      if not xtol: #non-iterables
        return cost(x)

    # set the range constraints
    xtol = asarray(xtol)
    bounds = zip( x - xtol, x + xtol )

    # optimize where initially x' = x
    stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(1)
    #XXX: edit settings?
    MINMAX = 1 #XXX: confirm MINMAX=1 is minimization
    if ipop: # use VTR
      results = diffev2(cost, bounds, ipop, ftol=ytol, gtol=None, \
                        itermon = stepmon, maxiter=imax, bounds=bounds, \
                        full_output=1, disp=0, handler=False)
    else: # use VTR
      results = fmin_powell(cost, x, ftol=ytol, gtol=None, \
                            itermon = stepmon, maxiter=imax, bounds=bounds, \
                            full_output=1, disp=0, handler=False)
   #solved = results[0]            # x'
    func_opt = MINMAX * results[1] # cost(x')
    if debug:
      print "solved: %s" % results[0]
      print "cost: %s" % func_opt

    # get the minimum distance |y - F(x')|
    return func_opt
  #########################################################################

  #XXX: better to do a single optimization rather than for each point ???
  d = [radius(model, point, ytol, xtol, ipop, imax) for point in target]
  return infeasibility(d, cutoff)


#def split_xy(params, npts):
#  """split params_{w,x,y} to params_{wx}, params_{y}
#  npts is [len(measure1),...,len(measureN)]  i.e. pm.pts"""
# #if not isinstance(npts,int):
#  _len = 2 * sum(npts)
#  return params[:_len], params[_len:]  
# #return params[:-npts], params[-npts:]

#def mend_xy(wx_param, y_param):
#  """append params_{y} to params_{wx} yielding params_{w,x,y}"""
#  param = wx_param[:]
#  param.extend(y_param)
#  return param


if __name__ == '__main__':
  ### conversions ###
  print "building a list of params(w,x,Y)..."
  pts = (2,2,2)
  param1 = [.5,.5,1,2, .25,.75,3,4, .125,.875,5,6, -1,-2,-3,-4,-5,-6,-7,-8]
  print "pts: %s" % str(pts)
  print "params: %s" % param1
 
  print "\nbuilding a scenario from the params..."
  # [store Y as 'values' OR register(F) for Y=F(X) OR points store y as 'val' ?]
  from mystic.math.dirac_measure import scenario
  pm = scenario()
  pm.load(param1, pts)
  print "pm.wts: %s" % str(pm.wts)
  print "pm.pos: %s" % str(pm.pos)
  W = pm.weights
  X = pm.coords
  Y = pm.values
  print "pm.weights: %s" % str(W)
  print "pm.coords: %s" % str(X)
  print "pm.values: %s" % str(Y)

  print "\nbuilding a dataset from the scenario..."
  # build a dataset (using X,Y)
  # [store W as 'weights' ?]
  from mystic.math.legacydata import dataset
  d = dataset()
  d.load(X, Y) 
  print "d.coords: %s" % str(d.coords) 
  print "d.values: %s" % str(d.values) 

  print "\nedit the dataset..."
  d[0].value = 0
  print "d.values: %s" % str(d.values) 
  # DON'T EDIT d[0].position... IT BREAKS PRODUCT MEASURE!!!

  print "\nupdate the scenario from the dataset..."
  # update pm(w,x,Y | W,X) from dataset(X,Y)
  pm.coords, pm.values = d.fetch()
  print "then, build a new list of params from the scenario..."
  # convert pm(w,x,Y | W,X) to params(w,x,Y)
  param1 = pm.flatten(all=True)
  print "params: %s" % str(param1)

  ### lipschitz ###
  param2 = [1.,0.,1,1, .75,.25,3,4, .5,.5,1,2, -8,-2,4,4,-3,-8,-8,0]
  qm = scenario()
  qm.load(param2, pts)
  L = [.25,.5,1.]

  print ""
 #print "param1: %s" % param1
 #print "param2: %s" % param2
  print "pm1.coords: %s" % pm.coords
  print "pm2.coords: %s" % qm.coords
  print "L = %s" % L

  print "manhattan distance..."
  print "pm1[0:3],pm2[0:3] =>\n", manhattan_distance(pm.coords[0:3],qm.coords[0:3]), "\n"
  print "pm1[0:1],pm2[0:3] =>\n", manhattan_distance(pm.coords[0:1],qm.coords[0:3]), "\n"
  print "pm1[0:1],pm2[0:1] =>\n", manhattan_distance(pm.coords[0:1],qm.coords[0:1]), "\n"

  print "pm1[0:3],pm2[0:3] values =>\n", manhattan_distance(pm.values[0:3],qm.values[0:3]), "\n"
  print "pm1[0:1],pm2[0:3] values =>\n", manhattan_distance(pm.values[0:1],qm.values[0:3]), "\n"
  print "pm1[0:1],pm2[0:1] values =>\n", manhattan_distance(pm.values[0:1],qm.values[0:1]), "\n"
   
  print "lipschitz metric..."
  print "pm1[0:3],pm2[0:3] =>\n", lipschitz_metric(L, pm.coords[0:3], qm.coords[0:3]), "\n"
  print "pm1[0:1],pm2[0:3] =>\n", lipschitz_metric(L, pm.coords[0:1], qm.coords[0:3]), "\n"
  print "pm1[0;1],pm2[0:1] =>\n", lipschitz_metric(L, pm.coords[0:1], qm.coords[0:1]), "\n"
  
  print "lipschitz distance..."
  print "(don't cutoff):\n%s" % lipschitz_distance(L, pm, qm, cutoff=None)
  print "from measures:\n%s" % lipschitz_distance(L, pm, qm)
  b = dataset()
  id = ['A','B','C','D','E','F','G','H']
  b.load(qm.coords, qm.values, ids=id) 
  print "from datasets:\n%s" % lipschitz_distance(L, d, b)
  print "from list of points:\n%s" % lipschitz_distance(L, d.raw, b.raw)
  print "individual points:\n%s" % lipschitz_distance(L, d.raw[0:1], b.raw[0:1])

  print ""
  print "is short:\n%s" % is_feasible( lipschitz_distance(L, pm, qm) )

  ### updates ###
  print "\nupdates to dataset..."
  print "original:\n%s" % b
  b.update(qm.coords, qm.values) 
  print "points from pm2:\n%s" % b
  b.update(pm.coords, pm.values) 
  print "points from pm1:\n%s" % b

  print "\nupdates to product measure..."
  print "orig: %s\n %s\n from %s" % (pm, pm.values, pm.flatten(all=True))
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, 0,-1,-2,-3,-4,-5,-6,-7]
  pm.update(par)
  print "alt: %s\n %s\n from %s" % (pm, pm.values, par)
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, -1,-2,-3,-4,-5,-6,-7,-8,-9]
  pm.update(par)
  print "alt2: %s\n %s\n from %s" % (pm, pm.values, par)
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, -2,-4,-6,-8]
  pm.update(par)
  print "alt3: %s\n %s\n from %s" % (pm, pm.values, par)
  par = [.5,.5,1,2, .25,.75,3,4]
  pm.update(par)
  print "alt4: %s\n %s\n from %s" % (pm, pm.values, par)

  ##### indexing #####
  b.update(pm.coords, pm.values)
  assert b[0].value == pm.values[0]
  assert b[1].value == pm.values[1]
  assert b[0].position == pm.coords[0] # slow
  assert b[0].position == pm.select(0, reduce=False)[0] # fast
  assert b[1].position == pm.coords[1] # slow
  assert b[1].position == pm.select(1) # fast
  assert pm.select(0,2) == [pm.select(0),pm.select(2)]

  # member calls #
  print "\ntesting mean_value..."
  print "mean_value: %s" % pm.get_mean_value()
  pm.set_mean_value(5.0)
  print "mean_value: %s" % pm.get_mean_value()

  print "\ntesting shortness, feasibility, validity..."
  assert pm.short_wrt_data(b) == True
  b.lipschitz = L
  assert pm.short_wrt_self(L) == False
  assert pm.short_wrt_data(b) == False
  pm.set_feasible(b)
  assert pm.short_wrt_self(L) == True
  assert pm.short_wrt_data(b) == True

  Cy = 0.1; Cx = 0.0
  model = lambda x:x[0]
  assert b.valid(model, ytol=Cy, xtol=Cx) == False
  assert pm.valid_wrt_model(model, ytol=Cy, xtol=Cx) == False
  pm.set_valid(model, cutoff=Cy, xtol=Cx)
  assert pm.valid_wrt_model(model, ytol=Cy, xtol=Cx) == True
  print "...done\n"


# EOF



#! /usr/bin/env python
"""
Classes for Dirac measure data objects.
Includes point, dirac_measure, product_measure, and scenario classes.
"""
# Adapted from seesaw2d.py in branches/UQ/math/examples2/ 
# For usage example, see seesaw2d_inf_example.py .

from mystic.math.measures import impose_mean, impose_expectation
from mystic.math.measures import impose_spread, impose_variance
from mystic.math.measures import impose_weight_norm

class point(object):
  """ 1-d object with weight and position

 queries:
  p.weight   --  returns weight
  p.position  --  returns position
  p.rms  --  returns the square root of sum of squared position

 settings:
  p.weight = w1  --  set the weight
  p.position = x1  --  set the position
"""

  def __init__(self, position, weight):
    self.weight = weight
    self.position = position
    return

  def __repr__(self):
    return "(%s @%s)" % (self.weight, self.position)

  def __rms(self): # square root of sum of squared positions
    from math import sqrt
    return sqrt(sum([i**2 for i in self.position]))

  # interface
  rms = property(__rms)

  pass


class dirac_measure(list):  #FIXME: meant to only accept points...
  """ a 1-d collection of points forming a 'discrete_measure'
  s = dirac_measure([point1, point2, ..., pointN])  
    where a point has weight and position

 queries:
  s.weights   --  returns list of weights
  s.coords  --  returns list of positions
  s.npts  --  returns the number of points
  s.mass  --  calculates sum of weights
  s.mean  --  calculates sum of weights*positions
  s.range  --  calculates |max - min| for positions
  s.var  --  calculates mean( |positions - mean(positions)|**2 )

 settings:
  s.weights = [w1, w2, ..., wn]  --  set the weights
  s.coords = [x1, x2, ..., xn]  --  set the positions
  s.normalize()  --  normalize the weights to 1.0
  s.mean(R)  --  set the mean
  s.range(R)  --  set the range
  s.var(R)  --  set the variance

 methods:
  s.get_expect(f)  --  calculate the expectation
  s.set_expect((center,delta), f)  --  impose expectation by adjusting positions

 notes:
  - constraints should impose that sum(weights) should be 1.0
  - assumes that s.n = len(s.coords) == len(s.weights)
"""

  def __weights(self):
    return [i.weight for i in self]

  def __positions(self):
    return [i.position for i in self]

  def __n(self):
    return len(self)

  def __mass(self):
    return sum(self.weights)
    #from mystic.math.measures import norm
    #return norm(self.weights)  # normalized by self.npts

  def __mean(self):
    from mystic.math.measures import mean
    return mean(self.coords, self.weights) 

  def __range(self):
    from mystic.math.measures import spread
    return spread(self.coords)

  def __variance(self):
    from mystic.math.measures import variance
    return variance(self.coords, self.weights)

  def __set_weights(self, weights):
    for i in range(len(weights)):
      self[i].weight = weights[i]
    return

  def __set_positions(self, positions):
    for i in range(len(positions)):
      self[i].position = positions[i]
    return

  def normalize(self):
    self.coords, self.weights = impose_weight_norm(self.coords, self.weights)
    return

  def __set_mean(self, m):
    self.coords = impose_mean(m, self.coords, self.weights)
    return

  def __set_range(self, r):
    self.coords = impose_spread(r, self.coords, self.weights)
    return

  def __set_variance(self, v):
    self.coords = impose_variance(v, self.coords, self.weights)
    return

  def get_expect(self, f):
    """calculate the expectation for a given function

Inputs:
    f -- a function that takes a list and returns a number
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?

    from mystic.math.measures import expectation
    coords = [(i,) for i in self.coords]
    return expectation(f, coords, self.weights)

  def set_expect(self, (m,D), f, bounds=None, constraints=None):
    """impose a expectation on a dirac measure

Inputs:
    (m,D) -- tuple of expectation m and acceptable deviation D
    f -- a function that takes a list and returns a number
    bounds -- tuple of lists of bounds  (lower_bounds, upper_bounds)
    constraints -- a function that takes a product_measure  c' = constraints(c)
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
#XXX: maybe also natural c' = constraints(c) where c is a dirac_measure ?

    if constraints:  # then need to adjust interface for 'impose_expectation'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    coords = impose_expectation((m,D), f, [self.npts], bounds, \
                                           self.weights,  constraints=cnstr) 
    from numpy import array
    self.coords = list(array(coords)[:,0])
   #from numpy import squeeze
   #self.coords = list(squeeze(coords))
    return

  # interface
  weights = property(__weights, __set_weights)
  coords = property(__positions, __set_positions)
  ###XXX: why not use 'points' also/instead?
  npts = property(__n )
  mass = property(__mass )
  range = property(__range, __set_range)
  mean = property(__mean, __set_mean)
  var = property(__variance, __set_variance)
  pass

class product_measure(list):  #FIXME: meant to only accept sets...
  """ a N-d product measure, a collection of dirac measures
  c = product_measure([measure1, measure2, ..., measureN])  
    where all measures are orthogonal

 queries:
  c.npts  --  returns total number of points
  c.weights   --  returns list of weights
  c.coords  --  returns list of position tuples
  c.mass  --  returns list of weight norms
  c.pts  --  returns number of points for each discrete measure
  c.wts  --  returns list of weights for each discrete measure
  c.pos  --  returns list of positions for each discrete measure

 settings:
  c.coords = [(x1,y1,z1),...]  --  set the positions (tuples in product measure)

 methods:
  c.pof(f)  --  calculate the probability of failure
  c.sampled_pof(f, npts) -- calculate the pof using sampled points
  c.get_expect(f)  --  calculate the expectation
  c.set_expect((center,delta), f)  --  impose expectation by adjusting positions
  c.flatten()  --  convert measure to a flat list of parameters
  c.load(params, pts)  --  'fill' the measure from a flat list of parameters
  c.update(params) -- 'update' the measure from a flat list of parameters

 notes:
  - constraints impose expect (center - delta) <= E <= (center + delta)
  - constraints impose sum(weights) == 1.0 for each set
  - assumes that c.npts = len(c.coords) == len(c.weights)
  - weight wxi should be same for each (yj,zk) at xi; similarly for wyi & wzi
"""
  def __val(self):
    raise NotImplementedError, "'value' is undefined in a dirac_measure"

  def __pts(self):
    return [i.npts for i in self]

  def __wts(self):
    return [i.weights for i in self]

  def __pos(self):
    return [i.coords for i in self]

  def __n(self):
    from numpy import product
    return product(self.pts)

  def __weights(self):
    from mystic.math.measures import _pack
    from numpy import product
    weights = _pack(self.wts)
    _weights = []
    for wts in weights:
      _weights.append(product(wts))
    return _weights

  def __positions(self):
    from mystic.math.measures import _pack
    return _pack(self.pos)

  def __set_positions(self, coords):
    from mystic.math.measures import _unpack
    coords = _unpack(coords, self.pts)
    for i in range(len(coords)):
      self[i].coords = coords[i]
    return

 #def __get_center(self):
 #  return self.__center

 #def __get_delta(self):
 #  return self.__delta

  def __mass(self):
    return [self[i].mass for i in range(len(self))]

  def get_expect(self, f):
    """calculate the expectation for a given function

Inputs:
    f -- a function that takes a list and returns a number
"""
    from mystic.math.measures import expectation
    return expectation(f, self.coords, self.weights)

  def set_expect(self, (m,D), f, bounds=None, constraints=None):
    """impose a expectation on a product measure

Inputs:
    (m,D) -- tuple of expectation m and acceptable deviation D
    f -- a function that takes a list and returns a number
    bounds -- tuple of lists of bounds  (lower_bounds, upper_bounds)
    constraints -- a function that takes a product_measure  c' = constraints(c)
"""
   #self.__center = m
   #self.__delta = D
    if constraints:  # then need to adjust interface for 'impose_expectation'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    self.coords = impose_expectation((m,D), f, self.pts, bounds, self.weights, \
                                                         constraints=cnstr) 
    return

  def pof(self, f):
    """calculate probability of failure over a given function, f,
where f takes a list of (product_measure) positions and returns a single value

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
"""
    u = 0
    set = zip(self.coords, self.weights)
    for x in set:
      if f(x[0]) <= 0.0:
        u += x[1]
    return u
  # for i in range(self.npts):
  #   #if f(self.coords[i]) > 0.0:  #NOTE: f(x) > 0.0 yields prob of success
  #   if f(self.coords[i]) <= 0.0:  #NOTE: f(x) <= 0.0 yields prob of failure
  #     u += self.weights[i]
  # return u  #XXX: does this need to be normalized?

  def sampled_pof(self, f, npts=10000):
    """calculate probability of failure over a given function, f,
where f takes a list of (product_measure) positions and returns a single value

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    npts -- number of points sampled from the underlying discrete measures
"""
    from mystic.math.samples import _pof_given_samples
    pts = self.support(npts)
    return _pof_given_samples(f, pts)

  def support(self, npts=10000):
    """randomly select support points from the underlying discrete measures

Inputs:
    npts -- number of points sampled from the underlying discrete measures

Returns:
    pts -- a nested list of len(prod_measure) lists, each of len(npts)
"""
    from mystic.math.measures import weighted_select as _select
    pts = []
    for i in range(npts):
      # for a single trial, select coords from all sets
      pts.append( [_select(set.coords, set.weights) for set in self] )

    # convert pts to len(prod_meas) lists, each of len(npts)
    from numpy import transpose
    return transpose(pts)  #XXX: assumes 'coords' is a list of floats

  def update(self, params):
    """update the product measure from a list of parameters

The dimensions of the product measure will not change"""
    pts = self.pts
    _len = 2 * sum(pts)

    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]

    pm = unflatten(params, pts)
    zo = pm.count([])
    self[:] = pm[:len(self) - zo] + self[len(pm) - zo:]
    return

  def load(self, params, pts):
    """load a list of parameters corresponding to N x 1D discrete measures

Inputs:
    params -- a list of parameters (see 'notes')
    pts -- number of points in each of the underlying discrete measures

Notes:
    To append len(pts) new discrete measures to product measure c, where
    pts = (M, N, ...)
    params = [wt_x1, ..., wt_xM, \
                 x1, ..., xM,    \
              wt_y1, ..., wt_yN, \
                 y1, ..., yN,    \
                     ...]
    Thus, the provided list is M weights and the corresponding M positions,
    followed by N weights and the corresponding N positions, with this
    pattern followed for each new dimension desired for the product measure.
"""
    _len = 2 * sum(pts)
    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]

    self.extend( unflatten(params, pts) )
    return

  def flatten(self):
    """flatten the product_measure into a list of parameters

Returns:
    params -- a list of parameters (see 'notes')

Notes:
    For a product measure c where c.pts = (M, N, ...), then
    params = [wt_x1, ..., wt_xM, \
                 x1, ..., xM,    \
              wt_y1, ..., wt_yN, \
                 y1, ..., yN,    \
                     ...]
    Thus, the returned list is M weights and the corresponding M positions,
    followed by N weights and the corresponding N positions, with this
    pattern followed for each dimension of the product measure.
"""
    params = flatten(self)
    return params

  #XXX: name stinks... better as "non_redundant"? ...is really a helper
  def differs_by_one(self, ith, all=True, index=True):
    """get the product measure coordinates where the associated binary
string differs by exactly one index

  Inputs:
    ith   = the target index
    all   = if False, return only the results for indicies < i
    index = if True, return the index of the results (not results themselves)
"""
    from mystic.math.compressed import index2binary, differs_by_one
    b = index2binary(range(self.npts), self.npts)
    return differs_by_one(ith, b, all, index) 

  def select(self, *index, **kwds):
    """generator for product measure coords due to selected position indicies
 (NOTE: only works for product measures of dimension 2^K)

  >>> r
  [[9, 8], [1, 3], [4, 2]]
  >>> r.select(*range(r.npts))
  [(9, 1, 4), (8, 1, 4), (9, 3, 4), (8, 3, 4), (9, 1, 2), (8, 1, 2), (9, 3, 2), (8, 3, 2)]
  >>>
  >>> _pack(r)
  [(9, 1, 4), (8, 1, 4), (9, 3, 4), (8, 3, 4), (9, 1, 2), (8, 1, 2), (9, 3, 2), (8, 3, 2)]
"""
    from mystic.math.compressed import index2binary, binary2coords
    v = index2binary(list(index), self.npts)
    return binary2coords(v, self.pos, **kwds)
    #XXX: '_pack' requires resorting ([::-1]) so that indexing is wrong.
    #     Better if modify mystic's pack to match sorting of binary strings ?

 #__center = None
 #__delta = None

  # interface
  npts = property(__n )
  weights = property(__weights )
  coords = property(__positions, __set_positions )
 #center = property(__get_center ) #FIXME: remove c.center and c.delta... or
 #delta = property(__get_delta )   #       replace with c._params (e.g. (m,D))
 #expect = property(__expect, __set_expect )
  mass = property(__mass )
  pts = property(__pts )
  wts = property(__wts )
  pos = property(__pos )
  pass


class scenario(product_measure):  #FIXME: meant to only accept sets...
  """ a N-d product measure (collection of dirac measures) with values
  s = scenario(product_measure, [value1, value2, ..., valueN])  
    where each point in the product measure is paried with a value
    (essentially, a dataset in product_measure representation)

 queries:
  s.npts  --  returns total number of points
  s.weights   --  returns list of weights
  s.coords  --  returns list of position tuples
  s.values  --  returns list of values
  s.mass  --  returns list of weight norms
  s.pts  --  returns number of points for each discrete measure
  s.wts  --  returns list of weights for each discrete measure
  s.pos  --  returns list of positions for each discrete measure

 settings:
  s.coords = [(x1,y1,z1),...]  --  set the positions (tuples in product measure)
  s.values = [v1,v2,v3,...]  --  set the values (correspond to position tuples)

 methods:
  s.pof(f)  --  calculate the probability of failure
  s.pof_value(f)  --  calculate the probability of failure using the values
  s.sampled_pof(f, npts) -- calculate the pof using sampled points
  s.get_expect(f)  --  calculate the expectation
  s.set_expect((center,delta), f)  --  impose expectation by adjusting positions
  s.get_mean_value()  --  calculate the mean values for a scenario
  s.set_mean_value(m)  --  impose mean value by adjusting values
  s.set_feasible(data)  --  impose shortness by adjusting positions and values
  s.short_wrt_data(data) -- check for shortness with respect to data
  s.short_wrt_self(L) -- check for shortness with respect to self
  s.set_valid(model) -- impose validity by adjusting positions and values
  s.valid_wrt_model(model) -- check for validity with respect to the model
  s.flatten()  --  convert measure to a flat list of parameters
  s.load(params, pts)  --  'fill' the measure from a flat list of parameters
  s.update(params) -- 'update' the measure from a flat list of parameters

 notes:
  - constraints impose expect (center - delta) <= E <= (center + delta)
  - constraints impose sum(weights) == 1.0 for each set
  - assumes that s.npts = len(s.coords) == len(s.weights)
  - weight wxi should be same for each (yj,zk) at xi; similarly for wyi & wzi
"""
  def __init__(self, pm=None, values=None):
    super(product_measure,self).__init__()
    if pm: 
      pm = product_measure(pm)
      self.load(pm.flatten(), pm.pts)
    if not values: values = []
    self.__Y = values # storage for values of s.coords
    return

  def __values(self):
    return self.__Y

  def __set_values(self, values):
    self.__Y = values[:]
    return

  def get_mean_value(self):  # get mean of y's
    """calculate the mean of the associated values for a scenario"""
    from mystic.math.measures import mean
    return mean(self.values, self.weights)

  def set_mean_value(self, m):  # set mean of y's
    """set the mean for the associated values of a scenario"""
    from mystic.math.measures import impose_mean
    self.values = impose_mean(m, self.values, self.weights)
    return

  def valid_wrt_model(self, model, blamelist=False, pairs=True, \
                                   all=False, raw=False, **kwds):
    """check for scenario validity with respect to the model

Inputs:
    model -- the model function, y' = F(x')
    blamelist -- if True, report which points are infeasible
    pairs -- if True, report indicies of infeasible points
    all -- if True, report results for each point (opposed to all points)
    raw -- if True, report numerical results (opposed to boolean results)

Additional Inputs:
    ytol -- maximum acceptable difference |y - F(x')|; a single value
    xtol -- maximum acceptable difference |x - x'|; an iterable or single value
    cutoff -- zero out distances less than cutoff; typically: ytol, 0.0, or None
    hausdorff -- norm; where if given, ytol = |y - F(x')| + |x - x'|/norm

Notes:
    xtol defines the n-dimensional base of a pilar of height ytol, centered at
    each point. The region inside the pilar defines the space where a "valid"
    model must intersect. If xtol is not specified, then the base of the pilar
    will be a dirac at x' = x. This function performs an optimization for each
    x to find an appropriate x'. While cutoff and ytol are very tightly related,
    they play a distinct role; ytol is used to set the optimization termination
    for an acceptable |y - F(x')|, while cutoff is applied post-optimization.
    If we are using the hausdorff norm, then ytol will set the optimization
    termination for an acceptable |y - F(x')| + |x - x'|/norm, where the x
    values are normalized by norm = hausdorff.
"""
    from mystic.math.legacydata import dataset 
    data = dataset() 
    data.load(self.coords, self.values)
   #data.lipschitz = L
    for i in range(len(data)):
      data[i].id = i
    return data.valid(model, blamelist=blamelist, pairs=pairs, \
                                       all=all, raw=raw, **kwds)

  def short_wrt_self(self, L, blamelist=False, pairs=True, \
                              all=False, raw=False, **kwds):
    """check for shortness with respect to the scenario itself

Inputs:
    L -- the lipschitz constant
    blamelist -- if True, report which points are infeasible
    pairs -- if True, report indicies of infeasible points
    all -- if True, report results for each point (opposed to all points)
    raw -- if True, report numerical results (opposed to boolean results)

Additional Inputs:
    tol -- maximum acceptable deviation from shortness
    cutoff -- zero out distances less than cutoff; typically: tol, 0.0, or None

Notes:
    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance 'tol', for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, cutoff and tol play distinct roles; tol is subtracted from
    calculation of the lipschitz_distance, while cutoff zeros out the value
    of any element less than the cutoff.
"""
    from mystic.math.legacydata import dataset 
    data = dataset() 
    data.load(self.coords, self.values)
    data.lipschitz = L
    for i in range(len(data)):
      data[i].id = i
    return data.short(blamelist=blamelist, pairs=pairs, \
                                           all=all, raw=raw, **kwds)

  def short_wrt_data(self, data, L=None, blamelist=False, pairs=True, \
                                         all=False, raw=False, **kwds):
    """check for shortness with respect to the given data

Inputs:
    data -- a collection of data points
    L -- the lipschitz constant, if different from that provided with data
    blamelist -- if True, report which points are infeasible
    pairs -- if True, report indicies of infeasible points
    all -- if True, report results for each point (opposed to all points)
    raw -- if True, report numerical results (opposed to boolean results)

Additional Inputs:
    tol -- maximum acceptable deviation from shortness
    cutoff -- zero out distances less than cutoff; typically cutoff = tol or 0.0

Notes:
    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance 'tol', for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, cutoff and tol play distinct roles; tol is subtracted from
    calculation of the lipschitz_distance, while cutoff zeros out the value
    of any element less than the cutoff.
"""
    from mystic.math.legacydata import dataset 
    _self = dataset() 
    _self.load(self.coords, self.values)
    _self.lipschitz = data.lipschitz
    for i in range(len(_self)):
      _self[i].id = i
    return _self.short(data, L=L, blamelist=blamelist, pairs=pairs, \
                                               all=all, raw=raw, **kwds)

  def set_feasible(self, data, cutoff=0.0, bounds=None, constraints=None, \
                                                  with_self=True, **kwds):
    """impose shortness on a scenario with respect to given data points

Inputs:
    data -- a collection of data points
    cutoff -- acceptable deviation from shortness

Additional Inputs:
    with_self -- if True, shortness will also be imposed with respect to self
    tol -- acceptable optimizer termination before sum(infeasibility) = 0.
    bounds -- a tuple of sample bounds:   bounds = (lower_bounds, upper_bounds)
    constraints -- a function that takes a flat list parameters
        x' = constraints(x)
"""
    # imposes: is_short(x, x'), is_short(x, z )
    # use additional 'constraints' kwds to impose: y >= m, norm(wi) = 1.0
    pm = impose_feasible(cutoff, data, guess=self.pts, bounds=bounds, \
                         constraints=constraints, with_self=with_self, **kwds)
    self.update( pm.flatten(all=True) )
    return

  def set_valid(self, model, cutoff=0.0, bounds=None, constraints=None, **kwds):
    """impose validity on a scenario with respect to given data points

Inputs:
    model -- the model function, y' = F(x'), that approximates reality, y = G(x)
    cutoff -- acceptable model invalidity |y - F(x')|

Additional Inputs:
    hausdorff -- norm; where if given, ytol = |y - F(x')| + |x - x'|/norm
    xtol -- acceptable pointwise graphical distance of model from reality
    tol -- acceptable optimizer termination before sum(infeasibility) = 0.
    bounds -- a tuple of sample bounds:   bounds = (lower_bounds, upper_bounds)
    constraints -- a function that takes a flat list parameters
        x' = constraints(x)

Notes:
    xtol defines the n-dimensional base of a pilar of height cutoff, centered at
    each point. The region inside the pilar defines the space where a "valid"
    model must intersect. If xtol is not specified, then the base of the pilar
    will be a dirac at x' = x. This function performs an optimization to find
    a set of points where the model is valid. Here, tol is used to set the
    optimization termination for the sum(graphical_distances), while cutoff is
    used in defining the graphical_distance between x,y and x',F(x').
"""
    # imposes is_feasible(R, Cv), where R = graphical_distance(model, pts)
    # use additional 'constraints' kwds to impose: y >= m, norm(wi) = 1.0
    pm = impose_valid(cutoff, model, guess=self, \
                      bounds=bounds, constraints=constraints, **kwds)
    self.update( pm.flatten(all=True) )
    return
 
  def pof_value(self, f):
    """calculate probability of failure over a given function, f,
where f takes a list of (scenario) values and returns a single value

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
"""
    u = 0
    set = zip(self.values, self.weights)
    for x in set:
      if f(x[0]) <= 0.0:
        u += x[1]
    return u

  def update(self, params): #XXX: overwritten.  create standalone instead ?
    """update the scenario from a list of parameters

The dimensions of the scenario will not change"""
    pts = self.pts
    _len = 2 * sum(pts)

    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]
      self.values = values[:len(self.values)] + self.values[len(values):] 

    pm = unflatten(params, pts)
    zo = pm.count([])
    self[:] = pm[:len(self) - zo] + self[len(pm) - zo:]
    return

  def load(self, params, pts): #XXX: overwritten.  create standalone instead ?
    """load a list of parameters corresponding to N x 1D discrete measures

Inputs:
    params -- a list of parameters (see 'notes')
    pts -- number of points in each of the underlying discrete measures

Notes:
    To append len(pts) new discrete measures to scenario c, where
    pts = (M, N, ...)
    params = [wt_x1, ..., wt_xM, \
                 x1, ..., xM,    \
              wt_y1, ..., wt_yN, \
                 y1, ..., yN,    \
                     ...]
    Thus, the provided list is M weights and the corresponding M positions,
    followed by N weights and the corresponding N positions, with this
    pattern followed for each new dimension desired for the scenario.
"""
    _len = 2 * sum(pts)
    if len(params)  >  _len:  # if Y-values are appended to params
      params, self.values  =  params[:_len], params[_len:]

    self.extend( unflatten(params, pts) )
    return

  def flatten(self, all=True): #XXX: overwritten.  create standalone instead ?
    """flatten the scenario into a list of parameters

Returns:
    params -- a list of parameters (see 'notes')

Notes:
    For a scenario c where c.pts = (M, N, ...), then
    params = [wt_x1, ..., wt_xM, \
                 x1, ..., xM,    \
              wt_y1, ..., wt_yN, \
                 y1, ..., yN,    \
                     ...]
    Thus, the returned list is M weights and the corresponding M positions,
    followed by N weights and the corresponding N positions, with this
    pattern followed for each dimension of the scenario.
"""
    params = flatten(self)
    if all: params.extend(self.values) # if Y-values, return those as well
    return params

  # interface
  values = property(__values, __set_values )
  pass


#---------------------------------------------
# creators and destructors from parameter list

def _mimic(samples, weights):
  """Generate a product_measure object from a list of N product measure
positions and a list of N weights. The resulting product measure will
mimic the original product measure's statistics, but be larger in size.

For example:
    >>> smp = [[-6,3,6],[-2,4],[1]]
    >>> wts = [[.4,.2,.4],[.5,.5],[1.]]
    >>> c = compose(samples, weights)
    >>> d = _mimic(c.coords, c.weights)
    >>> c[0].mean == d[0].mean
    True
    >>> c[1].range == d[1].range
    True
    >>> c.npts == d.npts
    False
    >>> c.npts == d[0].npts
    True
"""
  x = zip(*samples)                     # 'mimic' to a nested list
  w = [weights for i in range(len(x))]  # 'mimic' to a nested list
  return compose(x,w)


def _uniform_weights(samples):
  """generate a nested list of N x 1D weights from a nested list of N x 1D
discrete measure positions, where the weights have norm 1.0 and are uniform.

>>> c.pos
[[1, 2, 3], [4, 5], [6]]
>>> _uniform_weights(c.pos)
[[0.333333333333333, 0.333333333333333, 0.333333333333333], [0.5, 0.5], [1.0]]
"""
  from mystic.math.measures import normalize
  return [normalize([1.]*len(xi)) for xi in samples]


def _list_of_measures(samples, weights=None):
  """generate a list of N x 1D discrete measures from a nested list of N x 1D
discrete measure positions and a nested list of N x 1D weights.

Note this function does not return a product measure, it returns a list."""
  total = []
  if not weights: weights = _uniform_weights(samples)
  for i in range(len(samples)):
    next = dirac_measure()
    for j in range(len(samples[i])):
      next.append(point( samples[i][j], weights[i][j] ))
    total.append(next)
  return total


def compose(samples, weights=None):
  """Generate a product_measure object from a nested list of N x 1D
discrete measure positions and a nested list of N x 1D weights. If weights
are not provided, a uniform distribution with norm = 1.0 will be used."""
  if not weights: weights = _uniform_weights(samples)
  total = _list_of_measures(samples, weights)
  c = product_measure(total)
  return c


def decompose(c):
  """Decomposes a product_measure object into a nested list of
N x 1D discrete measure positions and a nested list of N x 1D weights."""
  from mystic.math.measures import _nested_split
  w, x = _nested_split(flatten(c), c.pts)
  return x, w


#def expand(data, npts):
#  """Generate a scenario object from a dataset. The scenario will have
#uniformly distributed weights and have dimensions given by pts."""
#  coords,values = data.fetch()
#  from mystic.math.measures import _unpack
#  pm = compose( _unpack(coords, npts) )
#  return scenario(pm, values[:pm.npts])


def unflatten(params, npts):
  """Map a list of random variables to N x 1D discrete measures
in a product_measure object."""
  from mystic.math.measures import _nested_split
  w, x = _nested_split(params, npts)
  return compose(x, w)


from itertools import chain #XXX: faster, but sloppy to have as importable
def flatten(c):
  """Flattens a product_measure object into a list."""
  rv = [(i.weights,i.coords) for i in c]
  # now flatten list of lists into just a list
  return list(chain(*chain(*rv))) # faster than mystic.tools.flatten


##### bounds-conserving-mean: borrowed from seismic/seismic.py #####
def bounded_mean(mean_x, samples, xmin, xmax, wts=None):
  from mystic.math.measures import impose_mean, impose_spread
  from mystic.math.measures import spread, mean
  from numpy import asarray
  a = impose_mean(mean_x, samples, wts)
  if min(a) < xmin:   # maintain the bound
    #print "violate lo(a)"
    s = spread(a) - 2*(xmin - min(a)) #XXX: needs compensation (as below) ?
    a = impose_mean(mean_x, impose_spread(s, samples, wts), wts)
  if max(a) > xmax:   # maintain the bound
    #print "violate hi(a)"
    s = spread(a) + 2*(xmax - max(a)) #XXX: needs compensation (as below) ?
    a = impose_mean(mean_x, impose_spread(s, samples, wts), wts)
  return asarray(a)
#####################################################################


#--------------------------------------------------
# constraints solvers and factories for feasibility

# used in self-consistent constraints function c(x) for
#   is_short(x, x') and is_short(x, z)
def norm_wts_constraintsFactory(pts):
  """factory for a constraints function that:
  - normalizes weights
"""
 #from dirac_measure import scenario
  def constrain(rv):
    "constrain:  sum(wi)_{k} = 1 for each k in K"
    pm = scenario()
    pm.load(rv, pts)      # here rv is param: w,x,y
    #impose: sum(wi)_{k} = 1 for each k in K
    norm = 1.0
    for i in range(len(pm)):
      w = pm[i].weights
      w[-1] = norm - sum(w[:-1])
      pm[i].weights = w
    rv = pm.flatten(all=True)
    return rv
  return constrain

# used in self-consistent constraints function c(x) for
#   is_short(x, x'), is_short(x, z), and y >= m
def mean_y_norm_wts_constraintsFactory(target, pts):
  """factory for a constraints function that:
  - imposes a mean on scenario values
  - normalizes weights
"""
 #from dirac_measure import scenario
  from mystic.math.measures import mean, impose_mean
 #target[0] is target mean
 #target[1] is acceptable deviation
  def constrain(rv):
    "constrain:  y >= m  and  sum(wi)_{k} = 1 for each k in K"
    pm = scenario()
    pm.load(rv, pts)      # here rv is param: w,x,y
    #impose: sum(wi)_{k} = 1 for each k in K
    norm = 1.0
    for i in range(len(pm)):
      w = pm[i].weights
      w[-1] = norm - sum(w[:-1])
      pm[i].weights = w
    #impose: y >= m 
    values, weights = pm.values, pm.weights
    y = float(mean(values, weights))
    if not (y >= float(target[0])):
      pm.values = impose_mean(target[0]+target[1], values, weights)
    rv = pm.flatten(all=True) 
    return rv
  return constrain

def impose_feasible(cutoff, data, guess=None, **kwds):
  """impose shortness on a given list of parameters w,x,y.

Optimization on w,x,y over the given bounds seeks sum(infeasibility) = 0.
  (this function is not ???-preserving)

Inputs:
    cutoff -- maximum acceptable deviation from shortness
    data -- a dataset of observed points (these points are 'static')
    guess -- the scenario providing an initial guess at feasibility,
        or a tuple of dimensions of the target scenario

Additional Inputs:
    tol -- acceptable optimizer termination before sum(infeasibility) = 0.
    bounds -- a tuple of sample bounds:   bounds = (lower_bounds, upper_bounds)
    constraints -- a function that takes a flat list parameters
        x' = constraints(x)

Outputs:
    pm -- a scenario with desired shortness
"""
  from numpy import sum, asarray
  from mystic.math.legacydata import dataset
  from my_paramtrans import lipschitz_distance, infeasibility, _npts
  if guess is None:
    message = "Requires a guess scenario, or a tuple of scenario dimensions."
    raise TypeError, message
  # get initial guess
  if hasattr(guess, 'pts'): # guess is a scenario
    pts = guess.pts    # number of x
    guess = guess.flatten(all=True)
  else:
    pts = guess        # guess is given as a tuple of 'pts'
    guess = None
  npts = _npts(pts)    # number of Y
  long_form = len(pts) - list(pts).count(2) # can use '2^K compressed format'

  # prepare bounds for solver
  bounds = kwds.pop('bounds', None)
  # if bounds are not set, use the default optimizer bounds
  if bounds is None:
    lower_bounds = []; upper_bounds = []
    for n in pts:  # bounds for n*x in each dimension  (x2 due to weights)
      lower_bounds += [None]*n * 2
      upper_bounds += [None]*n * 2
    # also need bounds for npts*y values
    lower_bounds += [None]*npts
    upper_bounds += [None]*npts
    bounds = lower_bounds, upper_bounds
  bounds = asarray(bounds).T

  # plug in the 'constraints' function:  param' = constraints(param)
  # constraints should impose_mean(y,w), and possibly sum(weights)
  constraints = kwds.pop('constraints', None) # default is no constraints
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  _self = kwds.pop('with_self', True) # default includes self in shortness
  if _self is not False: _self = True
  # tolerance for optimization on sum(y)
  tol = kwds.pop('tol', 0.0) # default
  npop = kwds.pop('npop', 20) #XXX: tune npop?
  maxiter = kwds.pop('maxiter', 1000) #XXX: tune maxiter?

  # if no guess was made, then use bounds constraints
  if guess is None:
    if npop:
      guess = bounds
    else:  # fmin_powell needs a list params (not bounds)
      guess = [(a + b)/2. for (a,b) in bounds]

  # construct cost function to reduce sum(lipschitz_distance)
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | sum(lipschitz_distance) | """
    _data = dataset()
    _pm = scenario()
    _pm.load(rv, pts)      # here rv is param: w,x,y
    if not long_form:
      coords = _pm.select(*range(npts))
    else: coords = _pm.coords
    _data.load( data.coords, data.values )                        # LOAD static
    if _self:
      _data.load( coords, _pm.values )                            # LOAD dynamic
    _data.lipschitz = data.lipschitz                              # LOAD L
    Rv = lipschitz_distance(_data.lipschitz, _pm, _data, tol=cutoff, **kwds)
    v = infeasibility(Rv, cutoff)
    return abs(sum(v))

  # construct and configure optimizer
  debug = False  #!!!
  maxfun = 1e+6
  crossover = 0.9; percent_change = 0.9
  ftol = abs(tol); gtol = None

  if debug:
    print "lower bounds: %s" % bounds.T[0]
    print "upper bounds: %s" % bounds.T[1]
  # print "initial value: %s" % guess
  # use optimization to get feasible points
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor
  from mystic.strategy import Best1Bin, Best1Exp
  evalmon = Monitor();  stepmon = Monitor(); strategy = Best1Exp
  if debug: stepmon = VerboseMonitor(10)  #!!!
  if npop: # use VTR
    results = diffev2(cost, guess, npop, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      cross=crossover, scale=percent_change, strategy=strategy,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  else: # use VTR
    results = fmin_powell(cost, guess, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  # repack the results
  pm = scenario()
  pm.load(results[0], pts)            # params: w,x,y
 #if debug: print "final cost: %s" % results[1]
  if debug and results[2] >= maxiter: # iterations
    print "Warning: constraints solver terminated at maximum iterations"
 #func_evals = results[3]           # evaluation
  return pm


def impose_valid(cutoff, model, guess=None, **kwds):
  """impose model validity on a given list of parameters w,x,y

Optimization on w,x,y over the given bounds seeks sum(infeasibility) = 0.
  (this function is not ???-preserving)

Inputs:
    cutoff -- maximum acceptable model invalidity |y - F(x')|; a single value
    model -- the model function, y' = F(x'), that approximates reality, y = G(x)
    guess -- the scenario providing an initial guess at validity,
        or a tuple of dimensions of the target scenario

Additional Inputs:
    hausdorff -- norm; where if given, ytol = |y - F(x')| + |x - x'|/norm
    xtol -- acceptable pointwise graphical distance of model from reality
    tol -- acceptable optimizer termination before sum(infeasibility) = 0.
    bounds -- a tuple of sample bounds:   bounds = (lower_bounds, upper_bounds)
    constraints -- a function that takes a flat list parameters
        x' = constraints(x)

Outputs:
    pm -- a scenario with desired model validity

Notes:
    xtol defines the n-dimensional base of a pilar of height cutoff, centered at
    each point. The region inside the pilar defines the space where a "valid"
    model must intersect. If xtol is not specified, then the base of the pilar
    will be a dirac at x' = x. This function performs an optimization to find
    a set of points where the model is valid. Here, tol is used to set the
    optimization termination for the sum(graphical_distances), while cutoff is
    used in defining the graphical_distance between x,y and x',F(x').
"""
  from numpy import sum as _sum, asarray
  from my_paramtrans import graphical_distance, infeasibility, _npts
  if guess is None:
    message = "Requires a guess scenario, or a tuple of scenario dimensions."
    raise TypeError, message
  # get initial guess
  if hasattr(guess, 'pts'): # guess is a scenario
    pts = guess.pts    # number of x
    guess = guess.flatten(all=True)
  else:
    pts = guess        # guess is given as a tuple of 'pts'
    guess = None
  npts = _npts(pts)    # number of Y

  # prepare bounds for solver
  bounds = kwds.pop('bounds', None)
  # if bounds are not set, use the default optimizer bounds
  if bounds is None:
    lower_bounds = []; upper_bounds = []
    for n in pts:  # bounds for n*x in each dimension  (x2 due to weights)
      lower_bounds += [None]*n * 2
      upper_bounds += [None]*n * 2
    # also need bounds for npts*y values
    lower_bounds += [None]*npts
    upper_bounds += [None]*npts
    bounds = lower_bounds, upper_bounds
  bounds = asarray(bounds).T

  # plug in the 'constraints' function:  param' = constraints(param)
  constraints = kwds.pop('constraints', None) # default is no constraints
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  # 'wiggle room' tolerances
  ipop = kwds.pop('ipop', 10) #XXX: tune ipop (inner optimization)?
  imax = kwds.pop('imax', 10) #XXX: tune imax (inner optimization)?
  # tolerance for optimization on sum(y)
  tol = kwds.pop('tol', 0.0) # default
  npop = kwds.pop('npop', 20) #XXX: tune npop (outer optimization)?
  maxiter = kwds.pop('maxiter', 1000) #XXX: tune maxiter (outer optimization)?

  # if no guess was made, then use bounds constraints
  if guess is None:
    if npop:
      guess = bounds
    else:  # fmin_powell needs a list params (not bounds)
      guess = [(a + b)/2. for (a,b) in bounds]

  # construct cost function to reduce sum(infeasibility)
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where: cost = | sum( infeasibility ) | """
    # converting rv to scenario
    points = scenario()
    points.load(rv, pts)
    # calculate infeasibility
    Rv = graphical_distance(model, points, ytol=cutoff, ipop=ipop, \
                                                        imax=imax, **kwds)
    v = infeasibility(Rv, cutoff)
    # converting v to E
    return _sum(v) #XXX: abs ?

  # construct and configure optimizer
  debug = False  #!!!
  maxfun = 1e+6
  crossover = 0.9; percent_change = 0.8
  ftol = abs(tol); gtol = None #XXX: optimally, should be VTRCOG...

  if debug:
    print "lower bounds: %s" % bounds.T[0]
    print "upper bounds: %s" % bounds.T[1]
  # print "initial value: %s" % guess
  # use optimization to get model-valid points
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor
  from mystic.strategy import Best1Bin, Best1Exp
  evalmon = Monitor();  stepmon = Monitor(); strategy = Best1Exp
  if debug: stepmon = VerboseMonitor(2)  #!!!
  if npop: # use VTR
    results = diffev2(cost, guess, npop, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      cross=crossover, scale=percent_change, strategy=strategy,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  else: # use VTR
    results = fmin_powell(cost, guess, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  # repack the results
  pm = scenario()
  pm.load(results[0], pts)            # params: w,x,y
 #if debug: print "final cost: %s" % results[1]
  if debug and results[2] >= maxiter: # iterations
    print "Warning: constraints solver terminated at maximum iterations"
 #func_evals = results[3]           # evaluation
  return pm


if __name__ == '__main__':
  from my_paramtrans import *
  model = lambda x:sum(x)
  a = [0,1,9,8, 1,0,4,6, 1,0,1,2, 0,1,2,3,4,5,6,7]
  feasability = 0.0; deviation = 0.01
  validity = 5.0; wiggle = 1.0
  y_mean = 5.0; y_buffer = 0.0
  L = [.75,.5,.25]
  bc = [(0,7,2),(3,0,2),(2,0,3),(1,0,3),(2,4,2)]
  bv = [5,3,1,4,8]
  pts = (2,2,2)
  from mystic.math.legacydata import dataset
  data = dataset()
  data.load(bc, bv)
  data.lipschitz = L
  pm = scenario()
  pm.load(a, pts)
  pc = pm.coords
  pv = pm.values
  #---
  _data = dataset()
  _data.load(bc, bv)
  _data.load(pc, pv)
  _data.lipschitz = data.lipschitz
  from numpy import sum
  ans = sum(lipschitz_distance(L, pm, _data))
  print "original: %s @ %s\n" % (ans, a)
 #print "pm: %s" % pm
 #print "data: %s" % data
  #---
  lb = [0,.5,-100,-100,  0,.5,-100,-100,  0,.5,-100,-100,   0,0,0,0,0,0,0,0]
  ub = [.5,1, 100, 100,  .5,1, 100, 100,  .5,1, 100, 100,   9,9,9,9,9,9,9,9]
  bounds = (lb,ub)

  _constrain = mean_y_norm_wts_constraintsFactory((y_mean,y_buffer), pts)
  results = impose_feasible(feasability, data, guess=pts, tol=deviation, \
                            bounds=bounds, constraints=_constrain)
  from mystic.math.measures import mean
  print "solved: %s" % results.flatten(all=True)
  print "mean(y): %s >= %s" % (mean(results.values, results.weights), y_mean)
  print "sum(wi): %s == 1.0" % [sum(w) for w in results.wts]

  print "\n---------------------------------------------------\n"

  bc = bc[:-2]
  ids = ['1','2','3']
  t = dataset()
  t.load(bc, map(model, bc), ids)
  t.update(t.coords, map(model, t.coords))
# r = dataset()
# r.load(t.coords, t.values)
# L = [0.1, 0.0, 0.0]
  print "%s" % t
  print "L: %s" % L
  print "shortness:"
  print lipschitz_distance(L, t, t, tol=0.0)
  
  print "\n---------------------------------------------------\n"

  print "Y: %s" % str(results.values)
  print "sum(wi): %s == 1.0" % [sum(w) for w in results.wts]


# EOF



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import sys

pkgdir = {'': 'python%s' % sys.version_info[0]}
VERSION = '0.9'

setup(name='httplib2',
        version=VERSION,
        author='Joe Gregorio',
        author_email='joe@bitworking.org',
        url='http://code.google.com/p/httplib2/',
        download_url='http://httplib2.googlecode.com/files/httplib2-%s.tar.gz' % VERSION,
        description='A comprehensive HTTP client library.',
        license='MIT',
        long_description="""

A comprehensive HTTP client library, ``httplib2`` supports many features left out of other HTTP libraries.

**HTTP and HTTPS**
  HTTPS support is only available if the socket module was compiled with SSL support.


**Keep-Alive**
  Supports HTTP 1.1 Keep-Alive, keeping the socket open and performing multiple requests over the same connection if possible.


**Authentication**
  The following three types of HTTP Authentication are supported. These can be used over both HTTP and HTTPS.

  * Digest
  * Basic
  * WSSE

**Caching**
  The module can optionally operate with a private cache that understands the Cache-Control:
  header and uses both the ETag and Last-Modified cache validators. Both file system
  and memcached based caches are supported.


**All Methods**
  The module can handle any HTTP request method, not just GET and POST.


**Redirects**
  Automatically follows 3XX redirects on GETs.


**Compression**
  Handles both 'deflate' and 'gzip' types of compression.


**Lost update support**
  Automatically adds back ETags into PUT requests to resources we have already cached. This implements Section 3.2 of Detecting the Lost Update Problem Using Unreserved Checkout


**Unit Tested**
  A large and growing set of unit tests.

        """,
        package_dir=pkgdir,
        packages=['httplib2'],
        package_data={'httplib2': ['*.txt']},
        classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
        ],
        )



from klepto import inf_cache as memoized
from klepto.archives import *
from klepto.keymaps import *

import numpy as np
data = np.arange(3)

arxiv = dir_archive('memo1',serialized=True,cached=True)
pik = picklemap(serializer='dill')
memo1 = memoized(cache=arxiv,keymap=pik)

@memo1
def testme1(x):
  return x

print(testme1(data))
print(testme1(data))

testme1.dump()

# THROWS  ValueError: I/O operation on closed file
print(testme1(data))





# To increment version
# git tag x.y.z
# git push --tags
# python setup.py sdist upload
from distutils.core import setup
setup(name = 'hickle',
      version = '1.3.2',
      description = 'Hickle - a HDF5 based version of pickle',
      author = 'Danny Price',
      author_email = 'dan@thetelegraphic.com',
      url = 'http://github.com/telegraphic/hickle',
      download_url='https://github.com/telegraphic/hickle/archive/1.3.2.tar.gz',
      platforms = 'Cross platform (Linux, Mac OSX, Windows)',
      keywords = ['pickle', 'hdf5', 'data storage', 'data export'],
      py_modules = ['hickle'],
      install_requires=['numpy', 'h5py']
      )


#! /usr/bin/env python
"""
... find the envelopes for graphical distance around a model

requires 2-D model, where one direction only has 3 discrete values
"""
from __future__ import with_statement
import dill as pickle
#try:
#  import simplejson as pickle
# #import json as pickle
#except ImportError:
#  pass

debug = False

from mystic.math.paramtrans import infeasibility, _get_xy

def graphical_distance(model, points, **kwds):
  """find the radius(x') that minimizes the graph between reality, y = G(x),
and an approximating function, y' = F(x')

Inputs:
  model = the model function, y' = F(x'), that approximates reality, y = G(x)
  points = object of type 'datapoint' to validate against; defines y = G(x)

Additional Inputs:
  ytol = maximum acceptable difference |y - F(x')|; a single value
  xtol = maximum acceptable difference |x - x'|; an iterable or single value
  cutoff = zero out distances less than cutoff; typically: ytol, 0.0, or None
  hausdorff = norm; where if given, ytol = |y - F(x')| + |x - x'|/norm

Returns:
  radius = minimum distance from x,G(x) to x',F(x') for each x

Notes:
  xtol defines the n-dimensional base of a pilar of height ytol, centered at
  each point. The region inside the pilar defines the space where a "valid"
  model must intersect. If xtol is not specified, then the base of the pilar
  will be a dirac at x' = x. This function performs an optimization for each
  x to find an appropriate x'. While cutoff and ytol are very tightly related,
  they play a distinct role; ytol is used to set the optimization termination
  for an acceptable |y - F(x')|, while cutoff is applied post-optimization.
  If we are using the hausdorff norm, then ytol will set the optimization
  termination for an acceptable |y - F(x')| + |x - x'|/norm, where the x
  values are normalized by norm = hausdorff.
""" #FIXME: update docs to show normalization in y
 #NotImplemented:
 #L = list of lipschitz constants, for use when lipschitz metric is desired
 #constraints = constraints function for finding minimum distance
  from mystic.math.legacydata import dataset
  from numpy import asarray, sum, isfinite, zeros, seterr
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor

  # ensure target xe and ye is a dataset
  target = dataset()
  target.load(*_get_xy(points))
  nyi = target.npts             # y's are target.values
  nxi = len(target.coords[-1])  # nxi = len(x) / len(y)
  
  # NOTE: the constraints function is a function over a single xe,ye
  #       because each underlying optimization is over a single xe,ye.
  #       thus, we 'pass' on using constraints at this time...
  constraints = None   # default is no constraints
  if kwds.has_key('constraints'): constraints = kwds.pop('constraints')
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  # get tolerance in y and wiggle room in x
  ytol = kwds.pop('ytol', 0.0)
  xtol = kwds.pop('xtol', 0.0) # default is to not allow 'wiggle room' in x 

  cutoff = ytol  # default is to zero out distances less than tolerance
  if kwds.has_key('cutoff'): cutoff = kwds.pop('cutoff')
  if cutoff is True: cutoff = ytol
  elif cutoff is False: cutoff = None
  ipop = kwds.pop('ipop', min(20, 3*nxi)) #XXX: tune ipop?
  imax = kwds.pop('imax', 1000) #XXX: tune imax?

  # get range for the dataset (normalization for hausdorff distance)
  hausdorff = kwds.pop('hausdorff', False)
  if not hausdorff:  # False, (), None, ...
    ptp = [0.0]*nxi
    yptp = 1.0
  elif hausdorff is True:
    from mystic.math.measures import spread
    ptp = [spread(xi) for xi in zip(*target.coords)]
    yptp = spread(target.values)
  else:
    try: #iterables
      if len(hausdorff) < nxi+1:
        hausdorff = list(hausdorff) + [0.0]*(nxi - len(hausdorff)) + [1.0]
      ptp = hausdorff[:-1]  # all the x
      yptp = hausdorff[-1]  # just the y
    except TypeError: #non-iterables
      ptp = [hausdorff]*nxi
      yptp = hausdorff

  #########################################################################
  def radius(model, point, ytol=0.0, xtol=0.0, ipop=None, imax=None):
    """graphical distance between a single point x,y and a model F(x')"""
    # given a single point x,y: find the radius = |y - F(x')| + delta
    # radius is just a minimization over x' of |y - F(x')| + delta
    # where we apply a constraints function (of box constraints) of
    # |x - x'| <= xtol  (for each i in x)
    #
    # if hausdorff = some iterable, delta = |x - x'|/hausdorff
    # if hausdorff = True, delta = |x - x'|/spread(x); using the dataset range
    # if hausdorff = False, delta = 0.0
    #
    # if ipop, then DE else Powell; ytol is used in VTR(ytol)
    # and will terminate when cost <= ytol
    x,y = _get_xy(point)
    y = asarray(y)

    # build the cost function
    if hausdorff: # distance in all directions
      def cost(rv):
        '''cost = |y - F(x')| + |x - x'| for each x,y (point in dataset)'''
        errs = seterr(invalid='ignore', divide='ignore') # turn off warning 
        z = abs((asarray(x) - rv)/ptp)  # normalize by range
        m = abs(y - model(rv))/yptp     # normalize by range
        m = m if isfinite(m) else 10.0**6
        seterr(invalid=errs['invalid'], divide=errs['divide']) # turn on warning
        return m + sum(z[isfinite(z)])
    else:  # vertical distance only
      def cost(rv):
        '''cost = |y - F(x')| for each x,y (point in dataset)'''
        return abs(y - model(rv))

    if debug:
      print "rv: %s" % str(x)
      print "cost: %s" % cost(x)

    # if xtol=0, radius is difference in x,y and x,F(x); skip the optimization
    try:
      if not imax or not max(xtol): #iterables
        return x, cost(x)
    except TypeError:
      if not xtol: #non-iterables
        return x, cost(x)

    # set the range constraints
    xtol = asarray(xtol)
    bounds = zip( x - xtol, x + xtol )

    if debug:
      print "lower: %s" % str(zip(*bounds)[0])
      print "upper: %s" % str(zip(*bounds)[1])

    # optimize where initially x' = x
    stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(1)
    #XXX: edit settings?
    MINMAX = 1 #XXX: confirm MINMAX=1 is minimization
    if ipop: # use VTR
      results = diffev2(cost, bounds, ipop, ftol=ytol, gtol=None, \
                        itermon = stepmon, maxiter=imax, bounds=bounds, \
                        full_output=1, disp=0, handler=False)
    else: # use VTR
      results = fmin_powell(cost, x, ftol=ytol, gtol=None, \
                            itermon = stepmon, maxiter=imax, bounds=bounds, \
                            full_output=1, disp=0, handler=False)
   #solved = results[0]            # x'
    func_opt = MINMAX * results[1] # cost(x')
    if debug:
      print "solved: %s" % results[0]
      print "cost: %s" % func_opt

    # get the minimum distance |y - F(x')|
    return results[0], func_opt
  #########################################################################

  #XXX: better to do a single optimization rather than for each point ???
  _d = [radius(model, point, ytol, xtol, ipop, imax) for point in target]
  r,d = zip(*_d)
  return asarray(r),infeasibility(d, cutoff)


def np_map(func, x): # *args
  "numpy map function"
  from numpy import vectorize
  vfunc = vectorize(func)
  return vfunc(x) # array of *args


def read_results(oldresults):
  try:
    with open(oldresults, 'rb') as file:
      memo = pickle.load(file)
  except:
    memo = {}
  return memo


def write_results(memo, newresults):
  with open(newresults, 'wb') as file:
    pickle.dump(memo, file)
  return


#def model_sausage(x):
#  from sausage import model_sausage
#  return model_sausage(x)

from memoize import memoized0_nopickle_round as memoized
# vectorize 'sausage_bounds' to find min/max for every evaluation point
@memoized(tol=3)
def model_sausage(x):
  "find the boundary of the sausage in Y for the given x"
  from sausage import sausage_bounds
  from runinfo import model, Y_bounds, settings
  ymin, yo, ymax = sausage_bounds(x, model, Y_bounds, **settings)
  from numpy import array
  return array([ymin, yo, ymax]).flatten()


if __name__ == '__main__':

  verbose = True #False
  oldresults = "cache.pkl"
  newresults = "cache.pkl"

  # get the run settings
  from runinfo import bounds, gridsize

  # get a queue of points to evaluate at
  from sausage import sausage_queue
  runs = sausage_queue(bounds, gridsize)
  if verbose: print "h v i:\n", runs

  # read any existing runs and results from archive
  model_sausage.memo.update(read_results(oldresults))
    
  # do a map of model_sausage over the runs
  # with 'serial' python
# results = map(model_sausage, runs)
  # with numpy
# results = np_map(model_sausage, runs)
  # with multiprocessing 
  from pathos.mp_map import mp_map as map  # multiprocessing
  results = map(model_sausage, runs)#, nproc=nodes)
  # with MPI
# nodes = 4 #len(runs)
# from pyina.ez_map import ez_map2 as map  # mpi4py
# results = map(model_sausage, runs, nnodes=nodes)
  # with MPI on a cluster
# nodes = '25:ppn=2'
# queue = 'weekdayQ'
# timelimit = '11:59:59'
# from pyina.ez_map import ez_map2 as map  # mpi4py
# from pyina.launchers import mpirun_launcher, torque_launcher
# from pyina.mappers import carddealer_mapper, equalportion_mapper
# from pyina.schedulers import torque_scheduler
# results = map(model_sausage, runs, nnodes=nodes, queue=queue, \
#               timelimit=timelimit, launcher=mpirun_launcher, \
#               scheduler=torque_scheduler)

  from numpy import asarray
  if verbose: print "yl yo yu:\n", asarray(results)

  # save the runs and results for later use
  write_results(model_sausage.memo, newresults)

### using tuples not arrays
##[memo.update({tuple(xi):tuple(yi)}) for (xi,yi) in zip(runs, results)]

  # check validity of results
  if verbose:
    assert model_sausage.memo == read_results(newresults)

  # linear interpolation to get envelope functions [6 func = 3 hi * 2 bounds]

  # apply envelopes as constraints to y




# EOF



#! /usr/bin/env python
# encoding: utf-8
"""
test_hickle.py
===============

Unit tests for hickle module.

"""

import os
from hickle import *
import unicodedata
import hashlib
import time

NESTED_DICT = {
        "level1_1" : {
            "level2_1" : [1, 2, 3],
            "level2_2" : [4, 5, 6]
        },
        "level1_2" : {
            "level2_1" : [1, 2, 3],
            "level2_2" : [4, 5, 6]
        },
        "level1_3" : {
            "level2_1" : {
                "level3_1" : [1, 2, 3],
                "level3_2" : [4, 5, 6]
            },
            "level2_2" : [4, 5, 6]
        }
    }

def test_string():
    """ Dumping and loading a string """
    filename, mode = 'test.h5', 'w'
    string_obj = "The quick brown fox jumps over the lazy dog"
    dump(string_obj, filename, mode)
    string_hkl = load(filename)
    #print "Initial list:   %s"%list_obj
    #print "Unhickled data: %s"%list_hkl
    try:
        assert type(string_obj) == type(string_hkl) == str
        assert string_obj == string_hkl
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        raise

def test_unicode():
    """ Dumping and loading a unicode string """
    filename, mode = 'test.h5', 'w'
    u = unichr(233) + unichr(0x0bf2) + unichr(3972) + unichr(6000)
    dump(u, filename, mode)
    u_hkl = load(filename)

    try:
        assert type(u) == type(u_hkl) == unicode
        assert u == u_hkl
        # For those interested, uncomment below to see what those codes are:
        # for i, c in enumerate(u_hkl):
        #     print i, '%04x' % ord(c), unicodedata.category(c),
        #     print unicodedata.name(c)
    except AssertionError:
        os.remove(filename)
        raise


def test_list():
    """ Dumping and loading a list """
    filename, mode = 'test.h5', 'w'
    list_obj = [1, 2, 3, 4, 5]
    dump(list_obj, filename, mode)
    list_hkl = load(filename)
    #print "Initial list:   %s"%list_obj
    #print "Unhickled data: %s"%list_hkl
    try:
        assert type(list_obj) == type(list_hkl) == list
        assert list_obj == list_hkl
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        raise

def test_set():
    """ Dumping and loading a list """
    filename, mode = 'test.h5', 'w'
    list_obj = set([1, 0, 3, 4.5, 11.2])
    dump(list_obj, filename, mode)
    list_hkl = load(filename)
    #print "Initial list:   %s"%list_obj
    #print "Unhickled data: %s"%list_hkl
    try:
        assert type(list_obj) == type(list_hkl) == set
        assert list_obj == list_hkl
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        raise

def test_numpy():
    """ Dumping and loading numpy array """
    filename, mode = 'test.h5', 'w'
    dtypes = ['float32', 'float64', 'complex64', 'complex128']
    
    for dt in dtypes:
        array_obj = np.ones(8, dtype=dt)
        dump(array_obj, filename, mode)
        array_hkl = load(filename)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        print array_hkl
        print array_obj
        raise


def test_masked():
    """ Test masked numpy array """
    filename, mode = 'test.h5', 'w'    
    a = np.ma.array([1,2,3,4], dtype='float32', mask=[0,1,0,0])
    
    dump(a, filename, mode)
    a_hkl = load(filename)
    
    try:
        assert a_hkl.dtype == a.dtype
        assert np.all((a_hkl, a))
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        print a_hkl
        print a
        raise
        

def test_dict():
    """ Test dictionary dumping and loading """
    filename, mode = 'test.h5', 'w'
    
    dd = {
        'name'   : 'Danny',
        'age'    : 28,
        'height' : 6.1,
        'dork'   : True,
        'nums'   : [1, 2, 3],
        'narr'   : np.array([1,2,3]),
        #'unic'   : u'dan[at]thetelegraphic.com'
    }
    
    
    dump(dd, filename, mode)
    dd_hkl = load(filename)
    
    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()
            
            if type(dd[k]) is type(np.array([1])):
                assert np.all((dd[k], dd_hkl[k]))
            else:
                #assert dd_hkl[k] == dd[k]
                pass
            assert type(dd_hkl[k]) == type(dd[k])
        except AssertionError:
            print k
            print dd_hkl[k]
            print dd[k]
            print type(dd_hkl[k]), type(dd[k])
            os.remove(filename)
            raise
    os.remove(filename)

def test_compression():
    """ Test compression on datasets"""
    
    filename, mode = 'test.h5', 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']
    
    comps = [None, 'gzip', 'lzf']
    
    for dt in dtypes:
        for cc in comps:
            array_obj = np.ones(32768, dtype=dt)
            dump(array_obj, filename, mode, compression=cc)
            print cc, os.path.getsize(filename)
            array_hkl = load(filename)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        print array_hkl
        print array_obj
        raise

def test_dict_int_key():
    """ Test for dictionaries with integer keys """
    filename, mode = 'test.h5', 'w'

    dd = {
        0: "test",
        1: "test2"
    }

    dump(dd, filename, mode)
    dd_hkl = load(filename)

    
    os.remove(filename)

def test_dict_nested():
    """ Test for dictionaries with integer keys """
    filename, mode = 'test.h5', 'w'

    dd = NESTED_DICT

    dump(dd, filename, mode)
    dd_hkl = load(filename)
    
    ll_hkl = dd_hkl["level1_3"]["level2_1"]["level3_1"]
    ll     = dd["level1_3"]["level2_1"]["level3_1"]
    assert ll == ll_hkl
    os.remove(filename)

def test_masked_dict():
    """ Test dictionaries with masked arrays """

    filename, mode = 'test.h5', 'w'

    dd = {
        "data"  : np.ma.array([1,2,3], mask=[True, False, False]),
        "data2" : np.array([1,2,3,4,5])
    }

    dump(dd, filename, mode)
    dd_hkl = load(filename)
    
    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()
            if type(dd[k]) is type(np.array([1])):
                assert np.all((dd[k], dd_hkl[k]))
            elif type(dd[k]) is type(np.ma.array([1])):
                print dd[k].data
                print dd_hkl[k].data
                assert np.allclose(dd[k].data, dd_hkl[k].data)
                assert np.allclose(dd[k].mask, dd_hkl[k].mask)
                
            assert type(dd_hkl[k]) == type(dd[k])

        except AssertionError:
            print k
            print dd_hkl[k]
            print dd[k]
            print type(dd_hkl[k]), type(dd[k])
            os.remove(filename)
            raise
    os.remove(filename)

def test_nomatch():
    """ Test for non-supported data types """
    filename, mode = 'nomatch.h5', 'w'

    dd = Exception('Nothing to see here')
    no_match = False
    dump(dd, filename, mode)
    
    #dd_hkl = load(filename)
    dd_hkl = load(filename, safe=False)
    
    assert type(dd_hkl) == type(dd) == Exception
    os.remove(filename)

def test_np_float():
    """ Test for singular np dtypes """
    filename, mode = 'np_float.h5', 'w'    
    
    dtype_list = (np.float16, np.float32, np.float64, 
                  np.complex64, np.complex128,
                  np.int8, np.int16, np.int32, np.int64,
                  np.uint8, np.uint16, np.uint32, np.uint64)
                  
    for dt in dtype_list:
    
        dd = dt(1)
        dump(dd, filename, mode)
        dd_hkl = load(filename)  
        assert dd == dd_hkl
        assert dd.dtype == dd_hkl.dtype
        os.remove(filename)

    dd = {}
    for dt in dtype_list:
        dd[str(dt)] = dt(1.0)
    dump(dd, filename, mode)
    dd_hkl = load(filename)

    #print dd
    for dt in dtype_list:
        assert dd[str(dt)] == dd_hkl[str(dt)]

    os.remove(filename)

def md5sum(filename, blocksize=65536):
    """ Compute MD5 sum for a given file """
    hash = hashlib.md5()
    with open(filename, "r+b") as f:
        for block in iter(lambda: f.read(blocksize), ""):
            hash.update(block)
    return hash.hexdigest()

def caching_dump(obj, filename, mode, **kwargs):
    """ Save arguments of all dump calls"""
    dump_cache.append((obj, filename, mode, kwargs))
    return hickle_dump(obj, filename, mode, **kwargs)

def test_track_times():
    """ Verify that track_times = False produces identical files"""
    hashes = []
    for obj, filename, mode, kwargs in dump_cache:
        kwargs['track_times'] = False
        hickle_dump(obj, filename, mode, **kwargs)
        hashes.append(md5sum(filename))
        os.remove(filename)

    time.sleep(1)

    for hash1, (obj, filename, mode, kwargs) in zip(hashes, dump_cache):
        hickle_dump(obj, filename, mode, **kwargs)
        hash2 = md5sum(filename)
        print hash1, hash2
        try:
            assert hash1 == hash2
            os.remove(filename)
        except AssertionError:
            os.remove(filename)
            raise


def test_comp_kwargs():
    """ Test compression with some kwargs for shuffle and chunking """

    filename, mode = 'test.h5', 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']

    comps = [None, 'gzip', 'lzf']
    chunks = [(100, 100), (250, 250)]
    shuffles = [True, False]
    scaleoffsets = [0, 1, 2]

    for dt in dtypes:
        for cc in comps:
            for ch in chunks:
                for sh in shuffles:
                    for so in scaleoffsets:
                        kwargs = {
                            'compression' : cc,
                            'dtype': dt,
                            'chunks': ch,
                            'shuffle': sh,
                            'scaleoffset': so
                        }
                        #array_obj = np.random.random_integers(low=-8192, high=8192, size=(1000, 1000)).astype(dt)
                        array_obj = NESTED_DICT
                        dump(array_obj, filename, mode, compression=cc)
                        print kwargs, os.path.getsize(filename)
                        array_hkl = load(filename)
    try:
        os.remove(filename)
    except AssertionError:
        os.remove(filename)
        print array_hkl
        print array_obj
        raise

def test_list_numpy():
    """ Test converting a list of numpy arrays """

    filename, mode = 'test.h5', 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = [a, b]

    dump(c, filename, mode)
    dd_hkl = load(filename)

    print dd_hkl

    assert isinstance(dd_hkl, list)
    assert isinstance(dd_hkl[0], np.ndarray)


    os.remove(filename)

def test_tuple_numpy():
    """ Test converting a list of numpy arrays """

    filename, mode = 'test.h5', 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = (a, b, a)

    dump(c, filename, mode)
    dd_hkl = load(filename)

    print dd_hkl

    assert isinstance(dd_hkl, tuple)
    assert isinstance(dd_hkl[0], np.ndarray)


    os.remove(filename)

def test_none():
    """ Test None type hickling """
    
    filename, mode = 'test.h5', 'w'

    a = None

    dump(a, filename, mode)
    dd_hkl = load(filename)
    print a
    print dd_hkl

    assert isinstance(dd_hkl, NoneType)

    os.remove(filename)
 
def test_dict_none():
     """ Test None type hickling """
    
     filename, mode = 'test.h5', 'w'

     a = {'a': 1, 'b' : None}

     dump(a, filename, mode)
     dd_hkl = load(filename)
     print a
     print dd_hkl

     assert isinstance(a['b'], NoneType)

     os.remove(filename)   
    
dump_cache = []
hickle_dump = dump
dump = caching_dump

if __name__ == '__main__':
  """ Some tests and examples"""
  test_dict_none()
  test_none()
  test_unicode()
  test_string()
  test_masked_dict()
  test_list()
  test_set()
  test_numpy()
  test_dict()
  test_compression()
  test_masked()
  test_dict_int_key()
  test_dict_nested()
  test_nomatch()
  test_np_float()
  test_track_times()
  time.sleep(2)
  test_comp_kwargs()
  test_list_numpy()
  test_tuple_numpy()
  
  print "ALL TESTS PASSED!"
  


# hickle: store any object in a HDF file
'''
hickle (especially, with dill) can be used to store single objects.

could store dicts of objects, however would need to load the entire dict...

could use directory w/ several HDF5 files as the dict, w/ one hickle per file
'''
d = {'x':3}
class e(object):
  pass
f = lambda x:x
g = e()

import hickle as hkl
import pickle
#NOTE: dill functionality was removed! WTF?
hkl.dump(d, 'test.hkl', 'w')
print hkl.load('test.hkl')

hkl.dump(e, 'test.hkl', 'w')
print pickle.loads(hkl.load('test.hkl').tostring())
#print hkl.load('test.hkl') #WTF: dill

#hkl.dump(f, 'test.hkl', 'w')
#print hkl.load('test.hkl') #WTF: dill

hkl.dump(g, 'test.hkl', 'w')
print pickle.loads(hkl.load('test.hkl').tostring())
#print hkl.load('test.hkl') #WTF: dill
